# A set of custom Ansible filter plugins used by DebOps

# Copyright (C) 2017-2019 Maciej Delmanowski <drybjed@gmail.com>
# Copyright (C) 2019      Robin Schneider <ypid@riseup.net>
# Copyright (C) 2017-2019 DebOps https://debops.org/


# This file is part of DebOps.
#
# DebOps is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DebOps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with DebOps. If not, see <https://www.gnu.org/licenses/>.

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)
from operator import itemgetter

__metaclass__ = type

try:
    unicode = unicode
except NameError:
    # 'unicode' is undefined, must be Python 3
    str = str
    unicode = str
    bytes = bytes
    basestring = (str, bytes)
else:
    # 'unicode' exists, must be Python 2
    str = str
    unicode = unicode
    bytes = str
    basestring = basestring


def _check_if_key_in_nested_dict(key, dictionary):
    if isinstance(dictionary, dict):
        for k, v in dictionary.items():
            if k == key:
                return True
            elif isinstance(v, dict):
                if _check_if_key_in_nested_dict(key, v):
                    return True
            elif isinstance(v, list):
                for d in v:
                    if _check_if_key_in_nested_dict(key, d):
                        return True

    return False


def _handle_copy_id_from(parsed_config, element, current_param):
    if 'copy_id_from' in element:
        if element['copy_id_from'] in parsed_config:
            id_src = element['copy_id_from']
            current_param['id'] = (
                int(parsed_config[id_src]['id']) +
                int(parsed_config[id_src].get('weight', 0)))


def _handle_weight(element, current_param):
    if 'weight' in element:
        current_param['weight'] = (
            int(element.get('weight',
                current_param.get('weight', 0))) +
            int(current_param.get('weight', 0)))


def _get_real_weight(current_param):
    return int(current_param['id']) + int(current_param['weight'])


def _parse_kv_value(current_data, new_data, data_index):
    """Parse the parameter values and merge
    with existing ones conditionally.
    """

    if 'value' in new_data:
        old_value = current_data.get('value')
        old_state = current_data.get('state', 'present')
        new_value = new_data['value']
        new_value_cast = new_data.get('value_cast', None)

        if (new_value is None or
                isinstance(new_value, (basestring, int, float, bool))):
            if (old_value is None or isinstance(old_value,
                                                (basestring, int,
                                                 float, bool, dict))):
                if new_value_cast in ['null', 'none', 'None']:
                    current_data['value'] = None
                elif new_value_cast in ['int', 'integer']:
                    current_data['value'] = int(new_value)
                elif new_value_cast in ['str', 'string']:
                    current_data['value'] = str(new_value)
                elif new_value_cast in ['bool', 'boolean']:
                    current_data['value'] = bool(new_value)
                elif new_value_cast == 'float':
                    current_data['value'] = float(new_value)
                else:
                    current_data['value'] = new_value

            # TODO(drybjed): This never evaluates to true.
            #  if (old_value is not None and old_state in ['comment'] and
            #          current_data['state'] != 'comment'):
            #      current_data['state'] = 'present'

        elif isinstance(new_value, list):
            if isinstance(old_value, dict):
                dict_value = current_data.get('value', {}).copy()
            else:
                dict_value = {}

            for element_index, element in enumerate(new_value):
                if isinstance(element, (basestring, int)):
                    dict_element = dict_value.get(element, {}).copy()
                    if not dict_element.get('name'):
                        dict_element.update({
                            'name': element,
                            'id': ((data_index * 10) + element_index),
                            'weight': dict_element.get('weight', 0),
                            'state': 'present'})

                        dict_element['real_weight'] = _get_real_weight(
                                dict_element)
                        dict_value[element] = dict_element
                        current_data['value'] = dict_value

                elif (isinstance(element, dict) and
                        element.get('param', element.get('name')) and
                        element.get('state', 'present') != 'ignore'):
                    element_name = element.get('param', element.get('name'))

                    for cursor in ([element_name]
                                   if isinstance(element_name,
                                                 (basestring, int))
                                   else element_name):
                        dict_element = dict_value.get(cursor, {}).copy()
                        dict_element.update({
                            'name': cursor,
                            'id': ((data_index * 10) + element_index),
                            'weight': int(dict_element.get('weight', 0)),
                            'state': element.get('state', 'present')
                        })

                        _handle_copy_id_from(dict_value, element, dict_element)

                        if 'weight' in element:
                            dict_element['weight'] = (
                                int(element.get('weight',
                                    dict_element.get('weight', 0))) +
                                int(dict_element.get('weight', 0)))

                        dict_element['real_weight'] = _get_real_weight(
                                dict_element)

                        # Include any unknown keys.
                        for key in element.keys():
                            if key not in ['name', 'state', 'id', 'weight',
                                           'real_weight', 'param',
                                           'copy_id_from']:
                                dict_element[key] = element.get(key)

                        dict_value.update({cursor: dict_element})
                        current_data.update({'value': dict_value})


def parse_kv_config(*args, **kwargs):
    """Return a parsed list of key/value configuration options

    Optional arguments:

        name
            string, name of the primary dictionary key used as an indicator to
            merge the related dictionaries together. If not specified, 'name'
            will be set as default.
    """
    name = kwargs.get('name', "name")
    input_args = []
    parsed_config = {}

    # Flatten the input list.
    for sublist in list(args):
        for item in sublist:
            input_args.append(item)

    for element_index, element in enumerate(input_args):

        if isinstance(element, (basestring)):

            # This is a simple string, let's make it a dictionary so that it
            # can be correctly processed.
            # We assume that the string should be a 'name' parameter.
            element = {name: element}

        if isinstance(element, dict):
            if (any(x in [name] for x in element) and
                    element.get('state', 'present') != 'ignore'):

                param_name = element.get(name)

                if element.get('state', 'present') == 'append':

                    # In append mode, don't create new config entries
                    if (parsed_config.get(param_name, {})
                            .get('state', 'present') == 'init'):
                        continue

                current_param = (parsed_config[param_name].copy()
                                 if param_name in parsed_config
                                 else {})

                if element.get('state', 'present') == 'append':
                    current_param['state'] = current_param.get(
                            'state', 'present')
                else:
                    current_param['state'] = (
                        element.get('state', current_param.get(
                            'state', 'present')))

                if (current_param['state'] == 'init' and
                    (element.get('state', 'present') != 'init' and
                        _check_if_key_in_nested_dict(
                        'value', current_param))):
                    current_param['state'] = 'present'

                current_param.update({
                    name: param_name,  # in case of a new entry
                    'id': int(current_param.get('id', (element_index * 10))),
                    'weight': int(current_param.get('weight', 0)),
                    'separator': element.get('separator',
                                             current_param.get('separator',
                                                               False)),
                    'section': element.get('section',
                                           current_param.get('section',
                                                             'unknown'))
                })

                _handle_copy_id_from(parsed_config, element, current_param)
                _handle_weight(element, current_param)

                current_param['real_weight'] = _get_real_weight(current_param)

                _parse_kv_value(current_param, element,
                                current_param.get('id'))

                if 'option' in element:
                    current_param['option'] = element.get('option')

                if 'comment' in element:
                    current_param['comment'] = element.get('comment')

                merge_keys = []
                if isinstance(kwargs.get('merge_keys'), list):
                    merge_keys.extend(kwargs.get('merge_keys'))

                if 'options' not in merge_keys:
                    merge_keys.append('options')

                for key_name in merge_keys:
                    if key_name in element:
                        current_options = current_param.get(key_name, [])
                        current_param[key_name] = parse_kv_config(
                            current_options + element.get(key_name),
                            merge_keys=merge_keys)

                # Include any unknown keys
                for unknown_key in element.keys():
                    if (unknown_key not in merge_keys
                        and unknown_key not in [name, 'state', 'id',
                                                'weight', 'real_weight',
                                                'separator', 'value',
                                                'comment', 'option',
                                                'section']):
                        current_param[unknown_key] = element.get(unknown_key)

                parsed_config.update({param_name: current_param})

            # These parameters are special and should not be interpreted
            # directly as configuration options
            elif not all(x in [name, 'option', 'state', 'comment',
                               'section', 'weight', 'value', 'copy_id_from']
                         for x in element):
                for key, value in element.items():
                    current_param = (parsed_config[key].copy()
                                     if key in parsed_config else {})
                    current_param.update({
                        name: key,
                        'state': 'present',
                        'id': int(current_param.get('id',
                                                    (element_index * 10))),
                        'weight': int(current_param.get('weight', 0)),
                        'section': current_param.get('section', 'unknown')
                    })

                    current_param['real_weight'] = _get_real_weight(
                            current_param)

                    _parse_kv_value(current_param,
                                    {'value': value},
                                    current_param.get('id'))

                    parsed_config.update({key: current_param})

    # Expand the dictionary of configuration options into a list,
    # and return sorted by weight.
    output = []
    for key, params in parsed_config.items():
        if isinstance(params.get('value'), dict):
            params['value'] = sorted(
                params.get('value').values(),
                key=itemgetter('real_weight')
            )

        output.append(params)

    return sorted(output, key=itemgetter('real_weight'))


def parse_kv_items(*args, **kwargs):
    """Return a parsed list of with_items elements
    Optional arguments:

        name
            string, name of the primary dictionary key used as an indicator to
            merge the related dictionaries together. If not specified, 'name'
            will be set as default.

        empty
            dictionary, keys are parameter names which might be empty, values
            are key name or list of key names, first key with a value other
            than None will be used as the specified parameter. Examples:
                empty={'some_param':  'other_param',
                       'empty_param': ['param2', 'param1']}

        defaults
            dictionary, keys are parameter names, values are default values to
            use when a parameter is not specified. Examples:
                      defaults={'some_param': 'default_value'}

        merge_keys
            list of keys in the dictionary that will be merged together. If not
            specified, 'options' key instances will be merged by default.
    """
    name = kwargs.get('name', "name")
    empty = kwargs.get('empty', {})
    defaults = kwargs.get('defaults', {})
    merge_keys = list(set(kwargs.get('merge_keys', set())))

    if 'options' not in merge_keys:
        merge_keys.append('options')

    input_args = []

    # Flatten the input list.
    for sublist in args:
        input_args.extend(sublist)

    parsed_config = {}

    for element_index, element in enumerate(input_args):

        if isinstance(element, dict):
            element_state = element.get('state', 'present')
        elif isinstance(element, (basestring)):

            # This is a simple string, let's make it a dictionary so that it
            # can be correctly processed.
            # We assume that the string should be a 'name' parameter.
            element = {name: element}
            element_state = 'present'

        if isinstance(element, dict):
            if (any(x in [name] for x in element) and
                    element_state != 'ignore'):

                param_name = element.get(name)

                if element_state == 'append':

                    # In append mode, don't create new config entries.
                    if (param_name not in parsed_config or
                        parsed_config[param_name].get('state',
                                                      'present') == 'init'):
                        continue

                current_param = (parsed_config[param_name].copy()
                                 if param_name in parsed_config
                                 else {})

                if element_state == 'append':
                    current_param['state'] = current_param.get(
                            'state', 'present')
                else:
                    current_param['state'] = (
                        element.get('state', current_param.get(
                            'state', 'present')))

                if (current_param['state'] == 'init' and
                        (element_state != 'init' and
                            _check_if_key_in_nested_dict(
                                'value', current_param))):
                    current_param['state'] = 'present'

                current_param.update({
                    name: param_name,  # in case of a new entry
                    'id': int(current_param.get('id', (element_index * 10))),
                    'weight': int(current_param.get('weight', 0)),
                    'separator': element.get('separator',
                                             current_param.get('separator',
                                                               False))
                })

                _handle_copy_id_from(parsed_config, element, current_param)
                _handle_weight(element, current_param)

                current_param['real_weight'] = _get_real_weight(current_param)

                if 'comment' in element:
                    current_param['comment'] = element.get('comment')

                # Set any default keys defined for the filter.
                for k, v in defaults.items():
                    current_param[k] = current_param.get(k, v)

                for key_name in merge_keys:
                    if key_name in element:
                        current_options = current_param.get(key_name, [])
                        current_param[key_name] = parse_kv_config(
                            current_options + element.get(key_name),
                            merge_keys=merge_keys)

                known_keys = [name, 'state', 'id', 'weight',
                              'real_weight', 'separator',
                              'comment', 'options']

                # Include any unknown keys.
                for unknown_key in element.keys():
                    if (unknown_key not in known_keys and
                            unknown_key not in merge_keys):
                        current_param[unknown_key] = element.get(unknown_key)

                # Fill any empty keys using other keys.
                for key_to_set, keys_to_check in empty.items():
                    if key_to_set in current_param:
                        continue

                    for key_to_check in keys_to_check:
                        if key_to_check in current_param:
                            current_param[key_to_set] = \
                                    current_param[key_to_check]
                            break

                parsed_config[param_name] = current_param

    # Expand the dictionary of configuration options into a list,
    # and return sorted by weight.
    output = []
    for key, params in parsed_config.items():
        # FIXME: Make recursive.
        if isinstance(params.get('value'), dict):
            params['value'] = sorted(
                params.get('value').values(),
                key=itemgetter('real_weight')
            )

        output.append(params)

    return sorted(output, key=itemgetter('real_weight'))


class FilterModule(object):
    """Register custom filter plugins in Ansible"""

    def filters(self):
        return {
            'parse_kv_config': parse_kv_config,
            'parse_kv_items': parse_kv_items
        }


if __name__ == '__main__':
    import unittest
    import textwrap
    import yaml

    class Test(unittest.TestCase):

        def test_parse_kv_value_simple(self):
            current_data = yaml.safe_load(textwrap.dedent('''
            name: 'local'
            value: 'test'
            '''))

            new_data = yaml.safe_load(textwrap.dedent('''
            name: 'local'
            value: 'test2'
            '''))

            _parse_kv_value(current_data, new_data, 0)

            #  print(yaml.dump(current_data, default_flow_style=False))
            #  print(yaml.dump(expected_items, default_flow_style=False))

            self.assertEqual(current_data, new_data)

        def test_parse_kv_value_mixed(self):
            current_data = yaml.safe_load(textwrap.dedent('''
            name: 'local'
            value:
              - 'alpha'
              - 'test'
            '''))

            new_data = yaml.safe_load(textwrap.dedent('''
            name: 'local'
            value:
              - 'beta'
            '''))

            # FIXME: Not getting it. Why does the _parse_kv_value behave like
            # this?
            expected_data = yaml.safe_load(textwrap.dedent('''
            name: local
            value:
              beta:
                id: 0
                name: beta
                real_weight: 0
                state: present
                weight: 0
            '''))

            _parse_kv_value(current_data, new_data, 0)

            #  print(yaml.dump(current_data, default_flow_style=False))
            #  print(yaml.dump(new_data, default_flow_style=False))

            self.assertEqual(current_data, expected_data)

        def test_parse_kv_config(self):
            input_items = yaml.safe_load(textwrap.dedent('''
            - name: 'local'
              value: 'test'
            - name: 'local2'
              value: 'test2'
            - name: 'local'
              value: 'test3'
            - name: 'local_null'
              value: null
            '''))

            expected_items = yaml.safe_load(textwrap.dedent('''
            - id: 0
              name: local
              real_weight: 0
              section: unknown
              separator: false
              state: present
              value: test3
              weight: 0
            - id: 10
              name: local2
              real_weight: 10
              section: unknown
              separator: false
              state: present
              value: test2
              weight: 0
            - id: 30
              name: local_null
              real_weight: 30
              section: unknown
              separator: false
              state: present
              value: null
              weight: 0
            '''))

            items = parse_kv_config(input_items)

            #  print(yaml.dump(items, default_flow_style=False))
            #  print(yaml.dump(expected_items, default_flow_style=False))

            self.assertEqual(items, expected_items)

        def test_parse_kv_config_simple_string(self):
            input_items = yaml.safe_load(textwrap.dedent('''
            - 'simple_string'
            '''))

            expected_items = yaml.safe_load(textwrap.dedent('''
            - id: 0
              name: simple_string
              real_weight: 0
              section: unknown
              separator: false
              state: present
              weight: 0
            '''))

            items = parse_kv_config(input_items)

            #  print(yaml.dump(items, default_flow_style=False))
            #  print(yaml.dump(expected_items, default_flow_style=False))

            self.assertEqual(items, expected_items)

        def test_parse_kv_config_null_to_list(self):
            input_items = yaml.safe_load(textwrap.dedent('''
            - name: 'local'
              value: null
            - name: 'local'
              value: ['test1']
            '''))

            expected_items = yaml.safe_load(textwrap.dedent('''
            - id: 0
              name: local
              real_weight: 0
              section: unknown
              separator: false
              state: present
              value:
              - id: 0
                name: test1
                real_weight: 0
                state: present
                weight: 0
              weight: 0
            '''))

            items = parse_kv_config(input_items)

            #  print(yaml.dump(items, default_flow_style=False))
            #  print(yaml.dump(expected_items, default_flow_style=False))

            self.assertEqual(items, expected_items)

        def test_parse_kv_config_renamed(self):
            input_items = yaml.safe_load(textwrap.dedent('''
            - renamed: 'local'
              value: 'test'
            - renamed: 'local2'
              value: 'test2'
            - renamed: 'local'
              value: 'test3'
            '''))

            expected_items = yaml.safe_load(textwrap.dedent('''
            - id: 0
              renamed: local
              real_weight: 0
              section: unknown
              separator: false
              state: present
              value: test3
              weight: 0
            - id: 10
              renamed: local2
              real_weight: 10
              section: unknown
              separator: false
              state: present
              value: test2
              weight: 0
            '''))

            items = parse_kv_config(input_items, name='renamed')

            #  print(yaml.dump(items, default_flow_style=False))
            #  print(yaml.dump(expected_items, default_flow_style=False))

            self.assertEqual(items, expected_items)

        def test_parse_kv_items_simple_string(self):
            input_items = yaml.safe_load(textwrap.dedent('''
            - 'simple_string'
            '''))

            expected_items = yaml.safe_load(textwrap.dedent('''
            - id: 0
              name: simple_string
              real_weight: 0
              separator: false
              state: present
              weight: 0
            '''))

            items = parse_kv_items(input_items)

            #  print(yaml.dump(items, default_flow_style=False))
            #  print(yaml.dump(expected_items, default_flow_style=False))

            self.assertEqual(items, expected_items)

        def test_parse_kv_items_empty(self):
            input_items = yaml.safe_load(textwrap.dedent('''
            - name: 'name should be used as comment'
              options:

                - name: 'second level is ignored'
                  service: |-
                    second level is ignored so service will not become the
                    comment
                  value: 'test'
            '''))

            expected_items = yaml.safe_load(textwrap.dedent('''
            - comment: name should be used as comment
              id: 0
              name: name should be used as comment
              options:
              - id: 0
                name: second level is ignored
                real_weight: 0
                section: unknown
                separator: false
                service: |-
                    second level is ignored so service will not become the
                    comment
                state: present
                value: test
                weight: 0
              real_weight: 0
              separator: false
              state: present
              weight: 0
            '''))

            items = parse_kv_items(
                    input_items, empty={'comment': ['service', 'name']})

            #  print(yaml.dump(items, default_flow_style=False))
            #  print(yaml.dump(expected_items, default_flow_style=False))

            self.assertEqual(items, expected_items)

        def test_parse_kv_items_defaults(self):
            input_items = yaml.safe_load(textwrap.dedent('''
            - name: 'something'
              key1: 'existing'
              value: 'something'
            '''))

            expected_items = yaml.safe_load(textwrap.dedent('''
            - id: 0
              key1: existing
              key2: value2
              name: something
              real_weight: 0
              separator: false
              state: present
              value: something
              weight: 0
            '''))

            items = parse_kv_items(
                    input_items, defaults={'key1': 'value1', 'key2': 'value2'})

            #  print(yaml.dump(items, default_flow_style=False))
            #  print(yaml.dump(expected_items, default_flow_style=False))

            self.assertEqual(items, expected_items)

        def test_parse_kv_items_merge_keys(self):
            input_items = yaml.safe_load(textwrap.dedent('''
            - name: 'something'
              key1: 'existing'
              options:

                - name: 'nested1'
                  value: 'value1'

              test:

                - name: 'test_nested1'
                  value: 'test_value1'

            - name: 'something'
              options:

                - name: 'nested2'
                  value: 'value2'

              test:

                - name: 'test_nested2'
                  value: 'test_value2'
            '''))

            expected_items = yaml.safe_load(textwrap.dedent('''
            - id: 0
              key1: existing
              name: something
              options:
              - id: 0
                name: nested1
                real_weight: 0
                section: unknown
                separator: false
                state: present
                value: value1
                weight: 0
              - id: 10
                name: nested2
                real_weight: 10
                section: unknown
                separator: false
                state: present
                value: value2
                weight: 0
              test:
              - id: 0
                name: test_nested1
                real_weight: 0
                section: unknown
                separator: false
                state: present
                value: test_value1
                weight: 0
              - id: 10
                name: test_nested2
                real_weight: 10
                section: unknown
                separator: false
                state: present
                value: test_value2
                weight: 0
              real_weight: 0
              separator: false
              state: present
              weight: 0
            '''))

            items = parse_kv_items(input_items, merge_keys=['test'])

            #  print(yaml.dump(items, default_flow_style=False))
            #  print(yaml.dump(expected_items, default_flow_style=False))

            self.assertEqual(items, expected_items)

        def test_parse_kv_items_copy_id_from(self):
            input_items = yaml.safe_load(textwrap.dedent('''
            - name: 'second'
              weight: 10
              value: 'value1'

            - name: 'first'
              weight: -10
              copy_id_from: 'second'
              value: 'value2'
            '''))

            expected_items = yaml.safe_load(textwrap.dedent('''
            - copy_id_from: second
              id: 10
              name: first
              real_weight: 0
              separator: false
              state: present
              value: value2
              weight: -10
            - id: 0
              name: second
              real_weight: 10
              separator: false
              state: present
              value: value1
              weight: 10
            '''))

            items = parse_kv_items(input_items)

            #  print(yaml.dump(items, default_flow_style=False))
            #  print(yaml.dump(expected_items, default_flow_style=False))

            self.assertEqual(items, expected_items)

        def test_parse_kv_items(self):
            input_items1 = yaml.safe_load(textwrap.dedent('''
            - name: 'should-stay-init'
              options:

                - name: 'local'
                  value: 'test'

              state: 'init'


            - name: 'should-become-present'
              options:

                - name: 'local'
                  value: 'test'

              state: 'init'

            - name: 'should-become-present'
              options:

                - name: 'local'
                  value: 'test2'


            - name: 'should-become-present2'
              options:

                - name: 'local'
                  value: 'test'
                  state: 'init'

              state: 'init'

            - name: 'should-become-present2'
              options:

                - name: 'local'
                  value: 'test2'


            - name: 'should-become-present3'
              options:

                - name: 'local1'
                  comment: 'This comment should survive.'
                  options:

                    - name: 'local2'
                      value: 'test'
                      state: 'init'

                  state: 'init'

                - name: 'external1'
                  options:

                    - name: 'external2'
                      value: 'test'
                      state: 'init'

                  state: 'init'
            '''))

            input_items2 = yaml.safe_load(textwrap.dedent('''
            - name: 'should-become-present3'
              options:

                - name: 'local1'
                  options:

                    - name: 'local2'
                      value: 'test2'

                - name: 'external1'
                  options:

                    - name: 'external2'
                      value: 'test'
            '''))

            expected_items = yaml.safe_load(textwrap.dedent('''
            - id: 0
              name: should-stay-init
              options:
              - id: 0
                name: local
                real_weight: 0
                section: unknown
                separator: false
                state: present
                value: test
                weight: 0
              real_weight: 0
              separator: false
              state: init
              weight: 0
            - id: 10
              name: should-become-present
              options:
              - id: 0
                name: local
                real_weight: 0
                section: unknown
                separator: false
                state: present
                value: test2
                weight: 0
              real_weight: 10
              separator: false
              state: present
              weight: 0
            - id: 30
              name: should-become-present2
              options:
              - id: 0
                name: local
                real_weight: 0
                section: unknown
                separator: false
                state: present
                value: test2
                weight: 0
              real_weight: 30
              separator: false
              state: present
              weight: 0
            - id: 50
              name: should-become-present3
              options:
              - comment: This comment should survive.
                id: 0
                name: local1
                options:
                - id: 0
                  name: local2
                  real_weight: 0
                  section: unknown
                  separator: false
                  state: present
                  value: test2
                  weight: 0
                real_weight: 0
                section: unknown
                separator: false
                state: present
                weight: 0
              - id: 10
                name: external1
                options:
                - id: 0
                  name: external2
                  real_weight: 0
                  section: unknown
                  separator: false
                  state: present
                  value: test
                  weight: 0
                real_weight: 10
                section: unknown
                separator: false
                state: present
                weight: 0
              real_weight: 50
              separator: false
              state: present
              weight: 0
            '''))

            items = parse_kv_items(input_items1, input_items2)

            #  print(yaml.dump(items, default_flow_style=False))
            #  print(yaml.dump(expected_items, default_flow_style=False))

            self.assertEqual(items, expected_items)

        def test_parse_kv_items_renamed(self):
            input_items1 = yaml.safe_load(textwrap.dedent('''
            - renamed: 'should-stay-init'
              options:

                - name: 'local'
                  value: 'test'

              state: 'init'


            - renamed: 'should-become-present'
              options:

                - name: 'local'
                  value: 'test'

              state: 'init'

            - renamed: 'should-become-present'
              options:

                - name: 'local'
                  value: 'test2'


            - renamed: 'should-become-present2'
              options:

                - name: 'local'
                  value: 'test'
                  state: 'init'

              state: 'init'

            - renamed: 'should-become-present2'
              options:

                - name: 'local'
                  value: 'test2'


            - renamed: 'should-become-present3'
              options:

                - name: 'local1'
                  comment: 'This comment should survive.'
                  options:

                    - name: 'local2'
                      value: 'test'
                      state: 'init'

                  state: 'init'

                - name: 'external1'
                  options:

                    - name: 'external2'
                      value: 'test'
                      state: 'init'

                  state: 'init'
            '''))

            input_items2 = yaml.safe_load(textwrap.dedent('''
            - renamed: 'should-become-present3'
              options:

                - name: 'local1'
                  options:

                    - name: 'local2'
                      value: 'test2'

                - name: 'external1'
                  options:

                    - name: 'external2'
                      value: 'test'
            '''))

            expected_items = yaml.safe_load(textwrap.dedent('''
            - id: 0
              renamed: should-stay-init
              options:
              - id: 0
                name: local
                real_weight: 0
                section: unknown
                separator: false
                state: present
                value: test
                weight: 0
              real_weight: 0
              separator: false
              state: init
              weight: 0
            - id: 10
              renamed: should-become-present
              options:
              - id: 0
                name: local
                real_weight: 0
                section: unknown
                separator: false
                state: present
                value: test2
                weight: 0
              real_weight: 10
              separator: false
              state: present
              weight: 0
            - id: 30
              renamed: should-become-present2
              options:
              - id: 0
                name: local
                real_weight: 0
                section: unknown
                separator: false
                state: present
                value: test2
                weight: 0
              real_weight: 30
              separator: false
              state: present
              weight: 0
            - id: 50
              renamed: should-become-present3
              options:
              - comment: This comment should survive.
                id: 0
                name: local1
                options:
                - id: 0
                  name: local2
                  real_weight: 0
                  section: unknown
                  separator: false
                  state: present
                  value: test2
                  weight: 0
                real_weight: 0
                section: unknown
                separator: false
                state: present
                weight: 0
              - id: 10
                name: external1
                options:
                - id: 0
                  name: external2
                  real_weight: 0
                  section: unknown
                  separator: false
                  state: present
                  value: test
                  weight: 0
                real_weight: 10
                section: unknown
                separator: false
                state: present
                weight: 0
              real_weight: 50
              separator: false
              state: present
              weight: 0
            '''))

            items = parse_kv_items(input_items1, input_items2, name='renamed')

            #  print(yaml.dump(items, default_flow_style=False))
            #  print(yaml.dump(expected_items, default_flow_style=False))

            self.assertEqual(items, expected_items)

    unittest.main()



testinfra_hosts = ['ansible://other-host']


def test_proxy(host):
    with host.sudo():
        host.run("apt-get install -y curl")
    r = host.run("curl -s -m 5 http://other-host.$(hostname -d)/index.html")
    assert r.rc == 0
    assert 'SOMETHING' in r.stdout


def test_named_proxy(host):
    with host.sudo():
        host.run("apt-get install -y curl")
    for name in ('name1', 'name2'):
        r = host.run(f"curl -s -m 5 http://{name}.$(hostname -d)/index.html")
        assert r.rc == 0
        assert 'NAMEDBACKEND' in r.stdout



from bs4 import BeautifulSoup
from functools import lru_cache
import json
import re
import urllib.parse
from enough.common import retry
from enough.common import ansible_utils
import requests
import urllib3
import yaml
import testinfra
from tests.infrastructure import get_driver

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class IcingaHelper(object):

    icinga_host = 'icinga-host'

    # set by ../conftest.py
    @staticmethod
    def set_ansible_inventory(inventory):
        IcingaHelper.inventory = inventory

    @lru_cache(maxsize=1)
    def get_auth(self):
        host = testinfra.get_host(f'ansible://{IcingaHelper.icinga_host}',
                                  ansible_inventory=self.inventory)
        with host.sudo():
            f = host.file("/etc/icinga2/conf.d/api-users.conf")
            return (
                re.search('ApiUser "(.*)"', f.content_string).group(1),
                re.search('password = "(.*)"', f.content_string).group(1)
            )

    def get_web_session(self):
        p = ansible_utils.Playbook('.', '.', [self.inventory])
        username = list(p.get_role_variables('icinga2', 'icingaweb2_user').values())[0]
        password = list(p.get_role_variables('icinga2', 'icingaweb2_user_pass').values())[0]
        address = self.get_address()
        session = requests.Session()
        session.verify = 'certs'

        r = session.get(f'https://{address}/icingaweb2/authentication/login')
        r.raise_for_status()
        soup = BeautifulSoup(r.text, 'html.parser')
        CSRFToken = soup.select('#CSRFToken')[0]['value']

        r = session.post(f'https://{address}/icingaweb2/authentication/login',
                         allow_redirects=False,
                         data={'username': username, 'password': password, 'CSRFToken': CSRFToken,
                               'formUID': 'form_login', 'redirect': '', 'btn_submit': 'Login'})
        r.raise_for_status()
        return (session, address)

    @lru_cache(maxsize=1)
    def get_address(self):
        if get_driver() == 'openstack':
            vars_dir = f'{self.inventory}/group_vars/all'
            host = 'icinga.' + yaml.safe_load(
                open(vars_dir + '/domain.yml'))['domain']
        else:
            host = testinfra.get_host('ansible://icinga-service-group',
                                      ansible_inventory=self.inventory)
            host = host.ansible.get_variables()['ansible_host']
        return host

    def get_api_url(self):
        host = self.get_address()
        return f'https://{host}:5665/v1'

    def connection_params(self):
        return dict(
            headers={'Accept': 'application/json'},
            auth=self.get_auth(),
            timeout=5,
            verify=False,
        )

    def get_hosts(self, host=None):
        url = self.get_api_url()
        if host:
            url += f'/objects/hosts/{host}?attrs=name&attrs=state'
        else:
            url += f'/objects/hosts?attrs=name&attrs=state'

        answer = requests.get(url, **self.connection_params())
        answer.raise_for_status()
        results = answer.json().get('results', [])
        if host:
            assert len(results) == 1
        return {x['attrs']['name']: x['attrs']['state'] for x in results}

    @retry.retry((AssertionError, requests.exceptions.HTTPError), tries=7)
    def wait_for_service(self, name):
        self.schedule_check(name)

        name = urllib.parse.quote(name)
        url = f'{self.get_api_url()}/objects/services/{name}'
        answer = requests.get(url, **self.connection_params())
        answer.raise_for_status()

        results = answer.json().get('results', [])
        assert len(results) == 1

        attrs = results[0]['attrs']
        assert attrs['last_check'] > 0, f'{name} has not been checked yet'
        state = attrs['state']
        assert int(state) == 0, f"{name}: unexpected state ({state}: " \
            "{results[0]['attrs']})"
        return True

    def schedule_check(self, name):
        """force the check to reduce the waiting time"""
        data = {
            'type': 'Service',
            'filter': f'service.__name=="{name}"',
            'force': True,
        }
        url = f'{self.get_api_url()}/actions/reschedule-check'
        answer = requests.post(url, data=json.dumps(data),
                               **self.connection_params())

        answer.raise_for_status()
        results = answer.json().get('results', [])
        assert len(results) == 1
        assert int(results[0]['code']) == 200

    def is_service_ok(self, name):
        return self.wait_for_service(name)



from __future__ import print_function
from io import StringIO
import jinja2
import json
import logging
import os
import sys
import sh
import tempfile

from enough.version import __version__
from enough.common.retry import retry, RetryException

log = logging.getLogger(__name__)


class Docker(object):

    def __init__(self, **kwargs):
        self.root = kwargs.get('root', os.path.join(os.path.dirname(__file__), '..'))
        self.name = kwargs.get('name')
        self.port = kwargs.get('port', '8000')
        self.retry = kwargs.get('retry', 9)
        self.domain = kwargs.get('domain', 'enough.community')
        self.bake_docker(kwargs.get('docker'))
        self.bake_docker_compose(kwargs.get('docker_compose'))

    def bake_docker(self, docker):
        cmd = docker or 'docker'
        self.docker = sh.Command(cmd).bake(
            _truncate_exc=False,
            _tee=True,
            _tty_out=False,
            _out=lambda x: log.info(x.strip()),
            _err=lambda x: log.info(x.strip()),
        )

    def bake_docker_compose(self, docker_compose):
        content = self.get_compose_content()
        if log.getEffectiveLevel() <= logging.DEBUG:
            # do not use log.debug() because the output is unreadable on a single line
            # and multiline logging is bad practice
            print(content)
        compose = tempfile.NamedTemporaryFile()
        compose.write(content.encode('utf-8'))
        compose.flush()
        cmd = docker_compose or 'docker-compose'
        self.docker_compose = sh.Command(cmd).bake(
            '--project-name', 'enough',
            '--file', compose.name,
            _truncate_exc=False,
            _tee=True,
            _tty_out=False,
            _out=lambda x: log.info(x.strip()),
            _err=lambda x: log.info(x.strip()),
        )
        self.compose_file = compose  # so that it is kept until self is deleted

    def create_network(self, name):
        existing = self.docker.network.ls('--format={{ .Name }}',
                                          f'--filter=name={name}').stdout
        if existing != b'':
            return False
        self.docker.network.create(name)
        return True

    def get_compose_content(self):
        f = os.path.join(self.root, 'common/data/docker-compose.yml')
        return self.replace_content(open(f).read())

    def replace_content(self, content):
        return jinja2.Template(content).render(this=self)

    def get_ports(self):
        return str(self.port) + ':8000'

    def create_image(self):
        dockerfile = os.path.join(self.root, 'common/data/base.dockerfile')
        return self._create_image('base', '-f', dockerfile, '.')

    def _create_image(self, suffix, *args):
        name = self.get_image_name_with_version(suffix)
        build_args = ['--quiet', '--tag', name]
        self.docker.build(build_args + list(args))
        self.docker.tag(name, self.get_image_name(suffix))
        return name

    def get_image_name(self, suffix):
        if suffix:
            return self.name + '_' + suffix
        else:
            return self.name

    def get_image_name_with_version(self, suffix):
        return self.get_image_name(suffix) + ':' + str(__version__)

    def down(self):
        self.docker_compose('down')

    def up_wait_for_services(self):
        self.up()
        try:
            self.wait_for_services()
        except RetryException:
            self.print_logs()
            raise

    def create_or_update(self):
        if self.inspect('placeholder'):
            return False
        self.create_image()
        self.up_wait_for_services()
        return True

    def get_ip(self):
        return self.inspect('{{range .NetworkSettings.Networks}}{{.IPAddress}}{{end}}')[0]

    def get_public_port(self, private_port):
        bindings = self.inspect('{{ json .HostConfig.PortBindings }}')
        if bindings:
            bindings = json.loads(bindings[0])
            return bindings[f'{private_port}/tcp'][0]['HostPort']
        else:
            return None

    def up(self):
        self.docker_compose('up', '-d')

    def inspect(self, format):
        ids = self.docker_compose('ps', '-q', _iter=True)
        results = []
        for id in ids:
            id = id.strip()
            r = self.docker('inspect', f'--format={format}', id)
            results.append(r.stdout.decode('utf-8').strip())
        return results

    def wait_for_services(self):
        @retry(AssertionError, tries=self.retry)
        def wait():
            results = self.inspect('{{ .State.Health.Status }}')
            assert all([x.strip() == 'healthy' for x in results]), str(results)
            log.info("enough service " + self.name + " is healthy")
        wait()

    def get_logs(self):
        result = StringIO()
        print(f"docker compose logs {self.name}\n", file=result)
        self.docker_compose('logs', _out=result)
        print(str(self.inspect('{{ json .State.Health }}')))
        result.write(str(self.inspect('{{ json .State.Health }}')))
        return result.getvalue()

    def print_logs(self):
        print(self.get_logs(), file=sys.stderr)



from enough import settings
from allauth.socialaccount.models import SocialApp
from django.contrib.sites.models import Site
from django.core.management.base import BaseCommand

from enough.common.gitlab import GitLab


class Command(BaseCommand):
    help = 'Create an API user'

    def add_arguments(self, parser):
        parser.add_argument('--config', default=f'{settings.CONFIG_DIR}')
        parser.add_argument('domain')
        parser.add_argument('username')
        parser.add_argument('password')

    def setup_gitlab_auth(self, *args, **kwargs):
        provider = 'gitlab'
        name = 'GitLab'

        existing = SocialApp.objects.filter(provider=provider)
        if existing:
            return False

        gitlab = GitLab(f'https://lab.{kwargs["domain"]}')
        gitlab.login('root', kwargs['password'])
        gitlab.ensure_group_exists('enough', request_access_enabled=True, visibility='public')
        (client_id, client_secret) = gitlab.create_api_application(kwargs['domain'])

        a = SocialApp(provider=provider, name=name, secret=client_secret,
                      client_id=client_id, key='')
        a.save()

        # Now associate this provider with all site instances.
        sites = [i for i in Site.objects.all()]
        a.sites.add(*sites)

        return True

    def handle(self, *args, **kwargs):
        changed = False
        changed |= self.setup_gitlab_auth(*args, **kwargs)

        if changed:
            self.stdout.write('Changed')
        else:
            self.stdout.write('Up to date')



import requests
import yaml
import dns.resolver
import gitlab_utils
from urllib.parse import urlparse
from bs4 import BeautifulSoup

#
# debug with
#
# enough ssh api-host
# docker exec -ti enough_enough-enough_1 journalctl -n 200 -f --unit enough
#

testinfra_hosts = ['ansible://gitlab-host']


def get_domain(inventory):
    vars_dir = f'{inventory}/group_vars/all'
    return yaml.safe_load(open(vars_dir + '/domain.yml'))['domain']


def api_sign_in(pytestconfig, host):
    domain = get_domain(pytestconfig.getoption("--ansible-inventory"))

    #
    # GitLab home page
    #
    lab = requests.Session()
    lab_url = f'https://lab.{domain}'
    r = lab.get(lab_url + '/users/sign_in', verify='certs')
    r.raise_for_status()
    soup = BeautifulSoup(r.text, 'html.parser')
    authenticity_token = soup.select(
        'form[action="/users/sign_in"] input[name="authenticity_token"]')[0]['value']

    #
    # GitLab Login
    #
    r = lab.post(lab_url + '/users/sign_in', data={
        'authenticity_token': authenticity_token,
        'user[login]': 'root',
        'user[password]': gitlab_utils.get_password(),
        'user[remember_me]': 0,
    }, verify='certs')
    r.raise_for_status()

    #
    # Revoke all tokens (in case test is run multiple times)
    #
    r = lab.get(lab_url + '/profile/applications', verify='certs')
    r.raise_for_status()
    soup = BeautifulSoup(r.text, 'html.parser')
    for f in soup.select('.oauth-authorized-applications form'):
        data = {}
        for input in f.children:
            if input.name != 'input':
                continue
            data[input['name']] = input['value']
        url = urlparse(f['action'])
        r = lab.post(lab_url + url.path, params=url.query, data=data, verify='certs')
        r.raise_for_status()

    #
    # API login
    #
    api = requests.Session()
    api.url = f'https://api.{domain}'
    r = api.get(api.url + '/accounts/gitlab/login/?process=login',
                allow_redirects=False,
                verify='certs')
    # print(r.headers['Location'])
    assert 'oauth/authorize' in r.headers['Location']
    r.raise_for_status()

    #
    # GitLab OAuth confirmation page
    #
    r = lab.get(r.headers['Location'], verify='certs')
    r.raise_for_status()
    assert 'An error' not in r.text
    soup = BeautifulSoup(r.text, 'html.parser')
    data = {
        'commit': 'Authorize',
    }
    for input in soup.select('form[action="/oauth/authorize"]:nth-child(2) input[type="hidden"]'):
        if input.get('name') is None or input.get('value') is None:
            continue
        data[input['name']] = input['value']
    #  print(str(data))
    #  print(soup.prettify())
    r = lab.post(lab_url + '/oauth/authorize', data=data,
                 allow_redirects=False,
                 verify='certs')
    r.raise_for_status()
    assert 'accounts/gitlab/login/callback' in r.headers['Location']

    # print(r.headers['Location'])
    r = api.get(r.headers['Location'], allow_redirects=False, verify='certs')
    # print(r.headers['Location'])
    r.raise_for_status()
    if '/accounts/social/signup/' == r.headers['Location']:
        #
        # API confirm email
        #
        r = api.get(f"{api.url}{r.headers['Location']}", verify='certs')
        r.raise_for_status()
        soup = BeautifulSoup(r.text, 'html.parser')
        data = {}
        for input in soup.select('form input'):
            if input.get('name') is None or input.get('value') is None:
                continue
            data[input['name']] = input['value']
        # print(str(data))
        # print(soup.prettify())
        r = api.post(f'{api.url}/accounts/social/signup/', data=data,
                     allow_redirects=False, verify='certs')
        r.raise_for_status()

    assert r.headers['Location'] == '/member/'

    #
    # API member page
    #
    r = api.get(f"{api.url}{r.headers['Location']}", verify='certs')
    r.raise_for_status()
    #
    # Get rest-framework token
    #
    soup = BeautifulSoup(r.content, 'html.parser')
    # print(soup.prettify())
    element = soup.select('#token')
    assert len(element) == 1
    token = element[0].contents[0]
    assert len(token) == 40
    return token


#
# debug with
#
# enough ssh api-host
# docker exec -ti enough_enough-enough_1 journalctl -n 200 -f --unit enough
#
def test_delegate_test_dns(pytestconfig, host):
    token = api_sign_in(pytestconfig, host)
    domain = get_domain(pytestconfig.getoption("--ansible-inventory"))
    url = f"https://api.{domain}"
    s = requests.Session()
    s.headers = {'Authorization': f'Token {token}'}
    data = {
        "name": "foo",
        "ip": "1.2.3.4",
    }
    r = s.post(f'{url}/delegate-test-dns/', json=data, timeout=60, verify='certs')
    # print(r.text)
    r.raise_for_status()
    resolver = dns.resolver.Resolver()
    bind_ip = str(resolver.query(f'bind.{domain}.')[0])
    resolver.nameservers = [bind_ip]
    assert '1.2.3.4' == str(resolver.query(f'ns-foo.test.{domain}.', 'a')[0])


def test_delegate_dns(pytestconfig, host):
    domain = get_domain(pytestconfig.getoption("--ansible-inventory"))
    url = f"https://api.{domain}"
    s = requests.Session()
    data = {
        "name": f"bar",
        "ip": "5.6.7.8",
    }
    r = s.post(f'{url}/delegate-dns/', json=data, timeout=60, verify='certs')
    # print(r.text)
    r.raise_for_status()
    resolver = dns.resolver.Resolver()
    bind_ip = str(resolver.query(f'bind.{domain}.')[0])
    resolver.nameservers = [bind_ip]
    assert '5.6.7.8' == str(resolver.query(f'ns-bar.d.{domain}.', 'a')[0])


def test_ping(pytestconfig, host):
    domain = get_domain(pytestconfig.getoption("--ansible-inventory"))
    url = f"https://api.{domain}"
    s = requests.Session()
    r = s.get(f'{url}/ping/', timeout=60, verify='certs')
    r.raise_for_status()
    assert 'pong' == r.text



testinfra_hosts = ['ansible://bind-host']


def test_sshfp(host):
    domain = host.run('hostname -d').stdout.strip()
    #
    # expected side effect of
    #
    #     - role: install_ssh_records
    #       vars:
    #         install_ssh_records_host: lab
    #
    cmd = host.run('dig sshfp +noall +answer @ns1.{domain} lab.{domain}'.format(
        domain=domain))
    assert cmd.rc == 0
    assert 'IN SSHFP' in cmd.stdout
    assert 'lab.' + domain in cmd.stdout



import json
import logging

testinfra_hosts = ['ansible://infrastructure4-host']
logger = logging.getLogger(__name__)


def test_volume(host):
    cmd = host.run("""
    set -xe
    ! test -e /dev/mapper/spare
    """)
    logger.debug('stdout %s', cmd.stdout)
    logger.debug('stderr %s', cmd.stderr)
    assert 0 == cmd.rc

    lsblk = host.run('lsblk --tree --json')
    logger.debug('stdout %s', lsblk.stdout)
    logger.debug('stdout %s', lsblk.stderr)
    assert 0 == lsblk.rc
    blockdevices = json.loads(lsblk.stdout)['blockdevices']
    # check that there are two disks and one unused disk
    assert 2 == len([unused for unused in blockdevices if unused['type'] == 'disk'])
    assert 1 == len([unused for unused in blockdevices
                     if not unused.get('children', False) and unused['type'] == 'disk'])



from cliff.show import ShowOne
from cliff.command import Command

from enough import settings
from enough.common import Enough
from enough.common import options


class Create(ShowOne):
    "Create or update a host"

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument('name')
        return options.set_options(parser)

    def take_action(self, parsed_args):
        args = vars(self.app.options)
        args.update(vars(parsed_args))
        e = Enough(settings.CONFIG_DIR, settings.SHARE_DIR, **args)
        r = e.host.create_or_update()
        columns = ('name', 'user', 'port', 'ip')
        data = (parsed_args.name, 'debian', r['port'], r['ipv4'])
        return (columns, data)


class Delete(Command):
    "Delete a host"

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument('name', nargs='+')
        return options.set_options(parser)

    def take_action(self, parsed_args):
        args = vars(self.app.options)
        args.update(vars(parsed_args))
        e = Enough(settings.CONFIG_DIR, settings.SHARE_DIR, **args)
        e.host.delete()



def pytest_addoption(parser):
    parser.addoption(
        "--enough-hosts",
        action="store",
        default=("infrastructure1-host,infrastructure2-host,infrastructure3-host,"
                 "infrastructure4-host,infrastructure5-host"),
        help="list of hosts"
    )
    parser.addoption(
        "--enough-service",
        action="store",
        default="infrastructure",
        help="service"
    )



import logging
import os
import re
import requests

from enough import settings
from enough.common import openstack
from enough.common import dotenough
from enough.common import ansible_utils

log = logging.getLogger(__name__)


class Service(object):

    class NoHost(Exception):
        pass

    def __init__(self, config_dir, share_dir, **kwargs):
        self.args = kwargs
        self.config_dir = config_dir
        self.share_dir = share_dir
        self.ansible = ansible_utils.Ansible(
            self.config_dir, self.share_dir, kwargs.get('inventory', []))
        self.dotenough = dotenough.DotEnough(config_dir, self.args['domain'])
        self.dotenough.ensure()
        self.set_service_info()
        self.update_vpn_dependencies()

    def set_service_info(self):
        groups = self.ansible.get_groups()
        suffix = '-service-hosts'
        self.service2hosts = {
            name.replace(suffix, ''): hosts for name, hosts in groups.items()
            if name.endswith(suffix)
        }
        suffix = '-service-group'
        self.service2group = {
            name.replace(suffix, ''): hosts for name, hosts in groups.items()
            if name.endswith(suffix)
        }

    def ensure_non_empty_service_group(self):
        service = self.args['name']
        hosts = self.service2group.get(service)
        if not hosts and not self.args.get('host'):
            raise ServiceOpenStack.NoHost(
                f"service {service} needs a host, specify one with --host")
        if self.args.get('host'):
            self.dotenough.service_add_to_group(service, self.args['host'])
            self.set_service_info()
            hosts = self.service2group[service]
        return hosts

    def update_vpn_dependencies(self):
        hosts = self.ansible.ansible_inventory()['_meta']['hostvars'].keys()
        internal_hosts = set(self.hosts_with_internal_network(hosts))
        if not internal_hosts:
            return
        assert 'openvpn' in self.service2hosts
        openvpn_hosts = set(self.service2hosts['openvpn'])
        for service in self.service2hosts.keys():
            if service == 'openvpn':
                continue
            hosts = set(self.service2hosts[service])
            if internal_hosts & hosts:
                self.service2hosts[service] = list(internal_hosts | openvpn_hosts)

    def get_vpn_host(self):
        hosts = self.service2group.get('openvpn')
        if hosts:
            return hosts[0]
        else:
            return None

    def add_vpn_hosts_if_needed(self, hosts):
        internal_hosts = set(self.hosts_with_internal_network(hosts))
        if not internal_hosts:
            return hosts
        assert 'openvpn' in self.service2hosts
        openvpn_hosts = set(self.service2hosts['openvpn'])
        hosts = set(hosts)
        return list(hosts | openvpn_hosts)

    def hosts_with_internal_network(self, hosts):
        info = self.ansible.get_hostvars('network_internal_only', *hosts)
        return [host for (host, internal_only) in info.items() if internal_only is True]

    def service_from_host(self, host):
        found = None
        hosts_count = 0
        for service, hosts in self.service2hosts.items():
            if host in hosts:
                if hosts_count > 0 and len(hosts) > hosts_count:
                    continue
                found = service
                hosts_count = len(hosts)
        return found


class ServiceDocker(Service):

    def create_or_update(self):
        pass


class ServiceOpenStack(Service):

    class PingException(Exception):
        pass

    def __init__(self, config_dir, share_dir, **kwargs):
        super().__init__(config_dir, share_dir, **kwargs)
        self.args = kwargs
        self.dotenough = dotenough.DotEnoughOpenStack(config_dir, self.args['domain'])
        self.dotenough.ensure()

    def maybe_delegate_dns(self):
        subdomain_regexp = r'(.*)\.d\.(.*)'
        m = re.match(subdomain_regexp, self.args['domain'])
        if not m:
            log.debug(f'{self.args["domain"]} does not match "{subdomain_regexp}", '
                      'do not attempt to delegate the DNS')
            return False
        (subdomain, domain) = m.group(1, 2)
        api = f'api.{domain}'
        ping = f'https://{api}/ping/'
        r = requests.get(ping)
        if not r.ok:
            raise ServiceOpenStack.PingException(f'{ping} does not respond')

        h = openstack.Heat(self.config_dir, cloud=self.args['cloud'])
        s = openstack.Stack(self.config_dir,
                            h.get_stack_definition('bind-host'),
                            cloud=self.args['cloud'])
        s.set_public_key(self.dotenough.public_key())
        bind_host = s.create_or_update()
        r = requests.post(f'https://{api}/delegate-dns/',
                          json={
                              'name': subdomain,
                              'ip': bind_host['ipv4'],
                          })
        r.raise_for_status()
        return True

    def create_or_update(self):
        self.ensure_non_empty_service_group()
        hosts = self.service2hosts[self.args['name']]
        h = openstack.Heat(self.config_dir, cloud=self.args['cloud'])
        h.create_missings(hosts, self.dotenough.public_key())
        self.maybe_delegate_dns()
        playbook = ansible_utils.Playbook(self.config_dir, self.share_dir)
        if os.path.isabs(self.args["playbook"]):
            playbook_file = self.args["playbook"]
        else:
            playbook_file = f'{self.config_dir}/{self.args["playbook"]}'
            if not os.path.exists(playbook_file):
                playbook_file = f'{self.share_dir}/{self.args["playbook"]}'
        playbook.run([
            f'--private-key={self.dotenough.private_key()}',
            '--limit', ','.join(hosts + ['localhost']),
            playbook_file,
        ])
        return {'fqdn': f'{self.args["name"]}.{self.args["domain"]}'}


def service_factory(config_dir=settings.CONFIG_DIR, share_dir=settings.SHARE_DIR, **kwargs):
    if kwargs['driver'] == 'openstack':
        return ServiceOpenStack(config_dir, share_dir, **kwargs)
    else:
        return ServiceDocker(config_dir, share_dir, **kwargs)



import logging

testinfra_hosts = ['ansible://infrastructure1-host']
logger = logging.getLogger(__name__)


def test_encrypted_volume(host):
    cmd = host.run("""
    set -xe
    test -e /dev/mapper/spare
    grep -q /srv /etc/fstab
    test ! -e /srv/docker
    test ! -e /srv/snap
    """)
    logger.debug('stdout %s', cmd.stdout)
    logger.debug('stderr %s', cmd.stderr)
    assert 0 == cmd.rc



import os
from abc import ABC, abstractmethod
import shutil
import tempfile

from enough import settings
from enough.common import openstack
from enough.common import docker
from enough.common import dotenough
from enough.common import tcp


class Host(ABC):

    def __init__(self, config_dir, share_dir, **kwargs):
        self.config_dir = config_dir
        self.share_dir = share_dir
        self.args = kwargs

    @abstractmethod
    def create_or_update(self):
        pass

    @abstractmethod
    def delete(self):
        pass


class HostDocker(Host):

    class DockerInfrastructure(docker.Docker):

        def __init__(self, config_dir, **kwargs):
            self.config_dir = config_dir
            super().__init__(**kwargs)

        def create_image(self):
            name = super().create_image()
            dockerfile = os.path.join(self.root, 'internal/data/infrastructure.dockerfile')
            with tempfile.TemporaryDirectory() as d:
                shutil.copy(f'{self.config_dir}/infrastructure_key.pub', d)
                return self._create_image(None,
                                          '--build-arg', f'IMAGE_NAME={name}',
                                          '-f', dockerfile, d)

        def get_compose_content(self):
            f = os.path.join(self.root, 'internal/data/infrastructure-docker-compose.yml')
            return self.replace_content(open(f).read())

    def __init__(self, config_dir, share_dir, **kwargs):
        super().__init__(config_dir, share_dir)
        self.args = kwargs
        self.dotenough = dotenough.DotEnoughDocker(config_dir, self.args['domain'])
        self.dotenough.ensure()
        self.d = HostDocker.DockerInfrastructure(config_dir, **self.args)

    def create_or_update(self):
        self.d.create_network(self.args['domain'])
        self.d.name = self.args['name']
        port = self.d.get_public_port('22')
        if not port:
            port = tcp.free_port()
            self.args['port'] = port
            self.d = HostDocker.DockerInfrastructure(self.config_dir, **self.args)
            self.d.create_or_update()
        return {
            'ipv4': self.d.get_ip(),
            'port': '22',
        }

    def delete(self):
        domain = self.args['domain']
        for id in self.d.docker.ps('--filter', f'label=enough={domain}', '-q', _iter=True):
            self.d.docker.rm('-f', id.strip())


class HostOpenStack(Host):

    def __init__(self, config_dir, share_dir, **kwargs):
        super().__init__(config_dir, share_dir, **kwargs)
        self.args = kwargs
        self.dotenough = dotenough.DotEnoughOpenStack(config_dir, self.args['domain'])
        self.dotenough.ensure()

    def create_or_update(self):
        dotenough.Hosts(self.config_dir).ensure(self.args['name'])
        h = openstack.Heat(self.config_dir, **self.args)
        s = openstack.Stack(self.config_dir, h.get_stack_definition(self.args['name']), **self.args)
        s.set_public_key(f'{self.config_dir}/infrastructure_key.pub')
        return s.create_or_update()

    def delete(self):
        self.delete_hosts(self.args['name'])

    def delete_hosts(self, names):
        h = openstack.Heat(self.config_dir, **self.args)
        complete_deletion = []
        for name in names:
            s = openstack.Stack(self.config_dir, h.get_stack_definition(name), **self.args)
            if s.delete_no_wait():
                complete_deletion.append(name)
        for name in complete_deletion:
            s = openstack.Stack(self.config_dir, h.get_stack_definition(name), **self.args)
            s.delete_wait()


def host_factory(config_dir=settings.CONFIG_DIR, share_dir=settings.SHARE_DIR, **kwargs):
    if kwargs['driver'] == 'openstack':
        return HostOpenStack(config_dir, share_dir, **kwargs)
    else:
        return HostDocker(config_dir, share_dir, **kwargs)



import os
import shutil
import textwrap


def make_config_dir(domain, enough_dot_dir):
    os.environ['ENOUGH_DOT'] = str(enough_dot_dir)
    config_dir = f'{enough_dot_dir}/{domain}'
    if not os.path.exists(config_dir):
        os.makedirs(config_dir)
    return config_dir


def prepare_config_dir(domain, enough_dot_dir):
    os.environ['ENOUGH_DOMAIN'] = domain
    config_dir = make_config_dir(domain, enough_dot_dir)
    # Reuse existent SSH key. If it doesn't exist, then a key will be
    # generated by DotEnough.ensure_ssh_key.
    if os.path.isfile('infrastructure_key'):
        shutil.copy('infrastructure_key', f'{config_dir}/infrastructure_key')
        shutil.copy('infrastructure_key.pub', f'{config_dir}/infrastructure_key.pub')
    all_dir = f'{config_dir}/inventory/group_vars/all'
    if not os.path.exists(all_dir):
        os.makedirs(all_dir)
    shutil.copyfile('inventory/group_vars/all/clouds.yml', f'{all_dir}/clouds.yml')
    shutil.copyfile('inventory/group_vars/all/provision.yml', f'{all_dir}/provision.yml')
    open(f'{all_dir}/certificate.yml', 'w').write(textwrap.dedent(f"""\
    ---
    certificate_authority: letsencrypt_staging
    """))
    return config_dir



from enough.internal.cmd import main
from enough.version import __version__


def test_enough_install_script(capsys, mocker):
    # do not tamper with logging streams to avoid
    # ValueError: I/O operation on closed file.
    mocker.patch('cliff.app.App.configure_logging')
    assert main(['--debug', 'install', 'internal/data/install.sh']) == 0
    out, err = capsys.readouterr()
    assert 'docker run' in out
    assert 'function()' not in out
    assert __version__ in out


def test_enough_install_script_no_version(capsys, mocker):
    # do not tamper with logging streams to avoid
    # ValueError: I/O operation on closed file.
    mocker.patch('cliff.app.App.configure_logging')
    assert main(['--debug', 'install', 'internal/data/install.sh', '--no-version']) == 0
    out, err = capsys.readouterr()
    assert 'docker run' in out
    assert 'function()' not in out
    assert __version__ not in out


def test_enough_install_function(capsys, mocker):
    # do not tamper with logging streams to avoid
    # ValueError: I/O operation on closed file.
    mocker.patch('cliff.app.App.configure_logging')
    assert main(['--debug', 'install', '--function', 'internal/data/install.sh']) == 0
    out, err = capsys.readouterr()
    assert 'docker run' in out
    assert 'function enough()' in out
    assert __version__ in out


def test_enough_install_defaults(capsys, mocker):
    # do not tamper with logging streams to avoid
    # ValueError: I/O operation on closed file.
    mocker.patch('cliff.app.App.configure_logging')
    assert main(['--debug', 'install']) == 0
    out, err = capsys.readouterr()
    assert 'docker run' in out
    assert 'function enough()' in out
    assert __version__ in out


def test_enough_install_set_registry(capsys, mocker):
    # do not tamper with logging streams to avoid
    # ValueError: I/O operation on closed file.
    mocker.patch('cliff.app.App.configure_logging')
    registry = 'REGISTRY'
    assert main(['--debug', 'install', 'internal/data/install.sh', '--registry', registry]) == 0
    out, err = capsys.readouterr()
    assert 'docker run' in out
    assert 'function()' not in out
    assert registry in out


def test_enough_install_default_registry(capsys, mocker):
    # do not tamper with logging streams to avoid
    # ValueError: I/O operation on closed file.
    mocker.patch('cliff.app.App.configure_logging')
    assert main(['--debug', 'install', 'internal/data/install.sh']) == 0
    out, err = capsys.readouterr()
    assert 'docker run' in out
    assert 'function()' not in out
    assert 'enoughcommunity/' in out



import pytest
import os
import yaml
from enough.common import ansible_utils


from enough.common.dotenough import (
    Hosts,
    DotEnough,
    DotEnoughOpenStack,
    DotEnoughOpenStackUnknownProvider,
)


#
# Hosts
#
def test_hosts_create_delete(tmpdir):
    config_dir = str(tmpdir)
    h = Hosts(config_dir)
    host = 'HOST'
    ip = '1.2.3.4'
    port = '22'
    assert h.create_or_update(host, ip, port) is True
    assert h.create_or_update(host, ip, port) is False
    assert os.path.exists(f'{config_dir}/inventory/hosts.yml')
    h = Hosts(config_dir)
    assert h.create_or_update(host, ip, port) is False
    assert h.hosts[host]['ansible_host'] == ip
    assert h.hosts[host]['ansible_port'] == port

    assert h.missings([host, 'MISSING']) == ['MISSING']

    h.delete(host)
    h = Hosts(config_dir)
    assert h.hosts == {}


def test_hosts_ensure(tmpdir):
    config_dir = str(tmpdir)
    h = Hosts(config_dir)
    host = 'HOST'
    assert h.ensure(host) is True
    assert h.ensure(host) is False
    assert os.path.exists(f'{config_dir}/inventory/hosts.yml')
    h = Hosts(config_dir)
    assert h.ensure(host) is False
    assert h.hosts[host] == {}


#
# DotEnough
#
def test_service_add_to_group(tmpdir):
    d = DotEnough(tmpdir, 'test.com')
    service = 'SERVICE'
    group = DotEnough.service2group(service)
    host = 'HOST'
    expected = {
        group: {
            'hosts': {
                host: None,
            }
        }
    }
    assert d.service_add_to_group(service, host) == expected
    assert d.service_add_to_group(service, host) == expected
    other_host = 'OTHER'
    expected[group]['hosts'][other_host] = None
    assert d.service_add_to_group(service, other_host) == expected
    os.system(f'cat {tmpdir}/services.yml')


#
# DotEnoughOpenStack
#
@pytest.mark.parametrize("clouds,cloud,provider", [
    (
        {
            'clouds': {
                'production': {
                    'auth': {
                        'auth_url': 'https://identity.api.ams.fuga.cloud:443/v3'
                    }
                }
            }
        },
        'production', 'fuga'),
    (
        {
            'clouds': {
                'production': {
                    'auth': {
                        'auth_url': 'https://auth.cloud.ovh.net/v3/'
                    }
                }
            }
        },
        'production', 'ovh'),
])
def test_populate_provider_supported(tmpdir, clouds, cloud, provider):
    d = DotEnoughOpenStack(tmpdir, 'test.com')
    f = f'{d.config_dir}/inventory/group_vars/all/clouds.yml'
    open(f, 'w').write(yaml.dump(clouds))
    a = ansible_utils.Ansible(d.config_dir, d.config_dir,
                              ['tests/enough/common/test_init/provider'])
    f = d.populate_provider(a, 'production')
    v = yaml.safe_load(open(f).read())
    assert v['openstack_provider'] == provider


def test_populate_provider_set(tmpdir):
    d = DotEnoughOpenStack(tmpdir, 'test.com')
    f = f'{d.config_dir}/inventory/group_vars/all/openstack_provider.yml'
    open(f, 'w').write(yaml.dump({'openstack_provider': 'known'}))
    a = ansible_utils.Ansible(d.config_dir, d.config_dir,
                              ['tests/enough/common/test_init/provider'])
    assert d.populate_provider(a, 'production') == 'known'


def test_populate_provider_no_clouds(tmpdir):
    d = DotEnoughOpenStack(tmpdir, 'test.com')
    a = ansible_utils.Ansible(d.config_dir, d.config_dir,
                              ['tests/enough/common/test_init/provider'])
    assert d.populate_provider(a, 'production') is None


def test_populate_provider_unknown(tmpdir):
    d = DotEnoughOpenStack(tmpdir, 'test.com')
    clouds = {
        'clouds': {
            'production': {
                'auth': {
                    'auth_url': 'https://example.com/'
                }
            }
        }
    }
    f = f'{d.config_dir}/inventory/group_vars/all/clouds.yml'
    open(f, 'w').write(yaml.dump(clouds))
    a = ansible_utils.Ansible(d.config_dir, d.config_dir,
                              ['tests/enough/common/test_init/provider'])
    with pytest.raises(DotEnoughOpenStackUnknownProvider) as e:
        assert d.populate_provider(a, 'production') == ''
    assert 'example.com' in str(e)



import testinfra
from enough.common import retry

testinfra_hosts = ['ansible://weblate-host']


def test_weblate_send_mail(host):

    weblate_host = host
    postfix_host = testinfra.host.Host.get_host(
        'ansible://postfix-host',
        ansible_inventory=host.backend.ansible_inventory)

    with postfix_host.sudo():
        host.run("postsuper -d ALL")

    cmd = weblate_host.run("""
    cd /srv/weblate
    sudo docker-compose -f docker-compose-infrastructure.yml exec -T weblate weblate \
         sendtestemail loic-doomtofail@dachary.org
    """)
    print(cmd.stdout)
    print(cmd.stderr)
    assert 0 == cmd.rc

    @retry.retry(AssertionError, tries=5)
    def wait_for_mail():
        with postfix_host.sudo():
            cmd = postfix_host.run("""
            grep -q doomtofail /var/spool/postfix/hold/*
            """)
        print(cmd.stdout)
        assert cmd.rc == 0, f'{cmd.stdout} {cmd.stderr}'
    wait_for_mail()



testinfra_hosts = ['ansible://forum-host']


def test_discourse(host):
    cmd = host.run("""
    set -xe
    sudo apt-get install -y curl
    curl --silent https://forum.$(hostname -d) | \
       grep --quiet 'Congratulations, you installed Discourse!'
    """)
    print(cmd.stdout)
    print(cmd.stderr)
    assert 0 == cmd.rc



#!/usr/bin/env python

from setuptools import setup, find_packages
import os
import re
import subprocess

try:
    import pypandoc
    README = pypandoc.convert_file('README.md', 'rst')
except(IOError, ImportError):
    print('Warning: The "pandoc" support is required to convert '
          'the README.md to reStructuredText format')
    README = open('README.md').read()

try:
    unicode
except NameError:
    # Required for Python 3.x
    class unicode(object):
        def __new__(cls, s):
            if isinstance(s, str):
                return s
            return s and s.decode('utf-8') or None

SCRIPTS = [os.path.join('bin', n) for n in [
    'debops', 'debops-init', 'debops-task',
    'debops-defaults', 'debops-padlock', 'debops-update']]

# Retrieve the project version from 'git describe' command and store it in the
# VERSION file, needed for correct installation of the Python package
try:
    with open(os.devnull, 'w') as devnull:
        RELEASE = subprocess.check_output(
                ['git', 'describe'], stderr=devnull
                ).strip().lstrip(b'v')
    with open('VERSION', 'w') as version_file:
        version_file.write('{}\n'.format(RELEASE.decode('utf-8')))
except subprocess.CalledProcessError:
    try:
        RELEASE = open('VERSION').read().strip()
    except Exception:
        try:
            with file('CHANGELOG.rst') as changelog:
                for count, line in enumerate(changelog):
                    if re.search('^`debops v', line):
                        RELEASE = line.split()[1].rstrip(b'`_').lstrip(b'v')
                        break
            with open('VERSION', 'w') as version_file:
                version_file.write('{}\n'.format(RELEASE.decode('utf-8')))
        except Exception:
            RELEASE = '0.0.0'

try:
    # Symlink the 'ansible/' directory inside of the 'debops/' Python package
    # directory. The files will be included in the package using the
    # MANIFEST.in file. This requires 'python-setuptools' APT package from
    # 'jessie-backports' repository.
    if not os.path.exists('debops/ansible'):
        os.symlink('../ansible', 'debops/ansible')
    setup(
        name="debops",
        version=unicode(RELEASE),
        install_requires=['argparse', 'distro', 'future'],
        extras_require={
            'ansible': ['ansible', 'netaddr', 'passlib',
                        'python-ldap', 'dnspython', 'pyopenssl']
            },

        scripts=SCRIPTS,
        packages=find_packages(exclude=['tests']),
        include_package_data=True,

        # metadata for upload to PyPI
        author="DebOps Developers",
        author_email="debops-users@lists.debops.org",
        description="Your Debian-based data center in a box",
        long_description=README,
        license="GPL-3.0-only",
        keywords="ansible debian sysadmin",
        url="https://debops.org/",
        python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, '
                        '!=3.3.*, !=3.4.*, <4',
        download_url="https://github.com/debops/debops"
                    "/archive/v" + unicode(RELEASE) + ".tar.gz",
        classifiers=[
                    'Development Status :: 5 - Production/Stable',
                    'Environment :: Console',
                    'Intended Audience :: Information Technology',
                    'Intended Audience :: System Administrators',
                    'License :: OSI Approved :: GNU General Public License v3 '
                    'or later (GPLv3+)',
                    'Natural Language :: English',
                    'Operating System :: POSIX',
                    'Programming Language :: Other Scripting Engines',
                    'Programming Language :: Python',
                    'Programming Language :: Python :: 2',
                    'Programming Language :: Python :: 2.7',
                    'Programming Language :: Python :: 3',
                    'Programming Language :: Python :: 3.5',
                    'Programming Language :: Python :: 3.6',
                    'Topic :: System :: Installation/Setup',
                    'Topic :: System :: Systems Administration',
                    'Topic :: Utilities'
        ],

        # these are for easy_install (used by bdist_*)
        zip_safe=True
        #    entry_points = {
        #        "console_scripts": [
        #            "debops = debops.cmds.main:run",
        #        ],
        #    },
    )
finally:
    # Unlink the symlinked 'ansible/' directory
    if os.path.islink('debops/ansible'):
        os.unlink('debops/ansible')



from enough.common import retry
import sh


def test_localhost_vpn_route():
    sh.sudo('apt-get', 'install', '-y', 'openvpn')
    sh.sudo.openvpn('/etc/openvpn/localhost.conf',
                    _cwd='/etc/openvpn', _bg=True, _truncate_exc=False)

    @retry.retry(AssertionError, tries=5)
    def check_route():
        output = sh.ip.route()
        assert '10.30.20' in output

    check_route()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (C) 2016 Robin Schneider <ypid@riseup.net>
# Copyright (C) 2016 DebOps Project http://debops.org/
#
# debops-api is part of DebOps.
#
# debops-api is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3 of the
# License.
#
# debops-api is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import sys
import json
import re
import logging
import pprint
from distutils.version import StrictVersion
import shutil

import yaml
import git
from docutils import core
from docutils.writers.html4css1 import Writer, HTMLTranslator

__license__ = 'AGPL-3.0-only'
__author__ = 'Robin Schneider <ypid@riseup.net>'
__version__ = '0.1.0'

"""
debops-api - Machine readable metadata about the DebOps Project.
"""


class NoHeaderHTMLTranslator(HTMLTranslator):
    def __init__(self, document):
        HTMLTranslator.__init__(self, document)
        self.head_prefix = ['', '', '', '', '']
        self.body_prefix = []
        self.body_suffix = []
        self.stylesheet = []


def reSTify(string):
    _w = Writer()
    _w.translator_class = NoHeaderHTMLTranslator
    return core.publish_string(string, writer=_w)


class DebOpsAPI:

    def __init__(
        self,
        strict=True,
        docs_url_pattern=None,
        changelog_url_pattern=None,
        role_owner=None,
        test_mode=False,
    ):

        self._strict = strict
        self._docs_url_pattern = docs_url_pattern
        self._changelog_url_pattern = changelog_url_pattern
        self._role_owner = role_owner
        self._test_mode = test_mode

        self._metadata = {}
        self._roles = {}

    def _get_role_full_name(self, role_owner, role_name):
        """
        Return the Ansible role name according to Ansible Galaxy naming
        convention.
        """

        return '{}.{}'.format(role_owner, role_name)

    def _get_repo_url(self, role_owner, role_name):
        """
        Return repository URL for the given Ansible role.
        """

        github_base_url = 'https://github.com'
        return '/'.join([
            github_base_url,
            role_owner,
            'ansible-' + self._get_normalized_role_name(role_name),
        ])

    def _read_github_repos_api_response(
        self,
        api_res_encoded_json,
        role_owner=None
    ):
        """
        Read GitHub API response for a /users/:username/repos or
        /orgs/:org/repos query handed over as encoded JSON string.
        Unfinished because relying on GitHub API is discouraged for the DebOps
        Project in case there are other ways to do it.
        """

        api_response = json.load(api_res_encoded_json)
        for repo in api_response:
            _re = re.match(r'ansible-(?P<role_name>[a-z0-9_-]+)+$',
                           repo['name'])
            if not _re:
                continue

            role_name = _re.group('role_name')
            role_full_name = self._get_role_full_name(role_owner, role_name)
            role_owner = self._get_owner_from_vcs_url(repo['html_url'])

            metadata_from_api = {
                'role_owner': role_owner,
                'role_name': role_name,
                'vcs_url': repo['html_url'],
                'role_format_version': '0.1.0',
            }
            self._metadata.setdefault(role_full_name, {})
            self._metadata[role_full_name].update(metadata_from_api)

    def read_github_repos_api_file(self, file_path, role_owner=None):
        """
        Read GitHub API response file for a /users/:username/repos or
        /orgs/:org/repos query passed as file path.
        """

        with open(args.github_api_response_file) as gh_rsp_fh:
            self._read_github_repos_api_response(
                gh_rsp_fh,
                role_owner=role_owner,
            )

    def _interpret_role_dir_name(self, role_dir_name):
        """
        Extract and return information from the role directory name.
        """

        version_by_pattern_map = {
            '0.1.0': re.compile(
                r'^(?P<role_owner>[^.]+)\.(?P<role_name>[a-z0-9_-]+)\.rst$'),
            '0.2.0': re.compile(
                r'^ansible-(?P<role_name>[a-z0-9_-]+)$'),
        }

        for role_format_version, pattern in version_by_pattern_map.items():
            _re = pattern.search(role_dir_name)
            if _re:
                role_owner = None
                if 'role_owner' in _re.groups():
                    role_owner = _re.group('role_owner')
                role_name = _re.group('role_name')
                logger.debug('Detected docs format version {} '
                             'for owner: {}, name: {} from {}'.format(
                                    role_format_version,
                                    role_owner,
                                    role_name,
                                    role_dir_name,
                                ))
                return {
                    'role_format_version': role_format_version,
                    'role_owner': role_owner,
                    'role_name': role_name,
                }

        return None

    def _get_decoded_yaml(self, yaml_file_path):
        """
        Get decoded YAML file.
        """

        try:
            with open(yaml_file_path) as ansigenome_fh:
                return yaml.safe_load(ansigenome_fh)
        except OSError:
            return {}

    def _get_owner_from_vcs_url(self, vcs_url):
        """
        Return owner name from VCS URL.
        Return `ypid` for `https://github.com/ypid/ansible-packages/`.
        """

        _re = re.match('[^:]+://[^/]+/(?P<owner_name>[^/]+)/', vcs_url)
        if _re:
            owner_name = _re.group('owner_name')
            logger.debug("Detected owner '{}' for URL: {}".format(
                owner_name,
                vcs_url,
            ))
            return owner_name
        else:
            return None

    def _get_vcs_info(self, dir_path):
        """
        Read VCS metadata for the given directory path.
        """

        g = git.Git(dir_path)

        # %cd: committer date
        last_committer_date = g.log('-1', '--format=%cd', '--date=iso8601')
        #  logger.debug('Got last committer date {} for: {}'.format(
        #      last_committer_date,
        #      dir_path,
        #  ))

        #  describe_version = g.describe()
        #  logger.debug(describe_version)

        try:
            version = g.describe('--abbrev=0', '--tags')
        except Exception:
            # Did not work on Travis test.
            # except git.exc.GitCommandError:
            version = '0.0.0'

        if self._test_mode:
            # Fake committer date in test mode
            last_committer_date = '1970-01-01 00:00:00 +0000'

        metadata = {
            'vcs_last_committer_date': last_committer_date,
            'version': re.sub(r'^v', '', version),
        }

        if not self._test_mode and version != '0.0.0':
            try:
                commits_since_last_release = len(
                    g.log('{}...HEAD'.format(version), '--oneline').split('\n')
                )
            except Exception:
                commits_since_last_release = None

            if commits_since_last_release is not None:
                metadata.update({
                    'vcs_commits_since_last_release': (
                        commits_since_last_release),
                })

        return metadata

    def _get_maintainers_from_line(self, line):
        # Modeled with the natural language processing from AIML in mind.
        # TODO: Remove redundancy. Duplicated into ansigenome source code.
        # Origin: debops-api
        _re = re.match(
            r'^[^.]*?maintainers?[\W_]+(:?is|are)[\W_]+`?(?P<nicks>.+?)\.?$',
            line,
            re.IGNORECASE
        )
        if _re:
            return [x.rstrip('_') for x in re.split(r'[\s,]+',
                    _re.group('nicks')) if x not in ['and', ',']]
        else:
            return None

    def _get_maintainers_from_changelog(self, changes_file):
        # TODO: Remove redundancy. Duplicated into ansigenome source code.
        # Origin: debops-api
        """
        Extract the maintainer from CHANGES.rst file and return the nickname of
        the maintainer.
        """

        try:
            with open(changes_file, 'r') as changes_fh:
                for line in changes_fh:
                    nick = self._get_maintainers_from_line(line)
                    if nick is not None:
                        return nick
        except FileNotFoundError:
            return None
        return None

    def _get_role_metadata(self, role_path):
        """
        Read metadata for the given role.
        """

        role_metadata = {}

        role_metadata['ansigenome'] = self._get_decoded_yaml(
            os.path.join(role_path, 'meta', 'ansigenome.yml')
        )['ansigenome_info']

        role_metadata['meta'] = self._get_decoded_yaml(
            os.path.join(role_path, 'meta', 'main.yml')
        )

        maintainer_nicks = self._get_maintainers_from_changelog(
            os.path.join(role_path, 'CHANGES.rst')
        )
        if maintainer_nicks is not None:
            role_metadata['maintainer_nicks'] = maintainer_nicks
            role_metadata['role_format_version'] = '0.2.1'

        return role_metadata

    def _get_normalized_meta_ansigenome(self, meta_ansigenome):
        """
        Returns normalized meta/ansigenome.yml data intended for inclusion in
        self._metadata.
        """

        metadata = {}

        if 'authors' in meta_ansigenome:
            metadata.setdefault('authors', [])

            for author_item in meta_ansigenome['authors']:
                metadata['authors'].append({
                    'name': author_item['name'],
                    'nick': author_item['github'],
                    'maintainer': False,
                })

        return metadata

    def _get_normalized_role_name(self, role_name):
        """
        Returns normalized role name as used in URLs
        Example role name: `ansible`, returns: `role-ansible`.
        """

        if role_name == 'ansible':
            role_name = 'role-' + role_name
        return role_name

    def _get_normalized_meta_main(self, meta_main):
        """
        Returns normalized meta/main.yml data intended for inclusion in
        self._metadata.
        """

        metadata = {}
        if 'galaxy_info' not in meta_main:
            return metadata

        license_map = {
            'GNU General Public License v3': 'GPL-3.0-only',
        }
        skip_keys = [
            'company',
            'author',
        ]
        rename_keys = {
            'galaxy_tags': 'tags',
        }

        for k, v in meta_main['galaxy_info'].items():
            if k == 'license':
                if v in license_map:
                    v = license_map[v]
            if k in skip_keys:
                continue
            k = rename_keys.get(k, k)

            metadata[k] = v

        return metadata

    def gen_role_metadata(self):
        """
        Generate metadata based on already present metadata.
        """

        for role_full_name, metadata in self._metadata.items():

            role_owner = metadata['role_owner']
            role_name = metadata['role_name']

            additonal_metadata = {
                'normalized_role_name': (
                    self._get_normalized_role_name(role_name)),
                'ci_badge_url': (
                    'https://api.travis-ci.org/{}/ansible-{}.png'.format(
                        role_owner,
                        self._get_normalized_role_name(role_name),
                    )
                ),
                'ci_url': 'https://travis-ci.org/{}/ansible-{}'.format(
                    role_owner,
                    self._get_normalized_role_name(role_name),
                ),
                'test_suite_url': (
                    'https://github.com/debops/test-suite/'
                    'tree/master/ansible-{}'.format(
                        self._get_normalized_role_name(role_name),
                    )
                ),
                'galaxy_url': 'https://galaxy.ansible.com/{}/{}'.format(
                    role_owner,
                    role_name,
                ),
            }

            if 'vcs_url' in metadata:
                additonal_metadata.update({
                    'clone_url': metadata['vcs_url'] + '.git',
                    'issue_url': metadata['vcs_url'] + '/issues',
                    'pr_url': metadata['vcs_url'] + '/pulls',
                })

            if StrictVersion('0.2.0') <= \
                    StrictVersion(metadata['role_format_version']):

                if self._docs_url_pattern:
                    additonal_metadata['docs_url'] = (
                            self._docs_url_pattern.format(
                                role_owner=role_owner,
                                role_name=role_name,
                                normalized_role_name=(
                                    self._get_normalized_role_name(role_name)),
                            )
                        )

                if self._changelog_url_pattern:
                    additonal_metadata['changelog_url'] = (
                            self._changelog_url_pattern.format(
                                role_owner=role_owner,
                                role_name=role_name,
                                normalized_role_name=(
                                    self._get_normalized_role_name(role_name)),
                            )
                        )

            self._metadata[role_full_name].update(additonal_metadata)

    def read_role_metadata(self, role_path):
        """
        Read metadata from each role available in role_path.
        """

        for role_dir_name in os.listdir(role_path):
            role_dir_info = self._interpret_role_dir_name(role_dir_name)

            # Ignore roles with old docs format for now since it would be
            # required to get the meta data from external servers (and to
            # encourage conversion to the new docs format).
            if role_dir_info:
                if StrictVersion('0.2.0') <= StrictVersion(
                        role_dir_info['role_format_version']):
                    role_name = role_dir_info['role_name']
                    role_metadata = self._get_role_metadata(
                            os.path.join(role_path, role_dir_name))
                    role_vcs_url = role_metadata['ansigenome']['github_url']
                    role_owner = self._get_owner_from_vcs_url(role_vcs_url)
                    role_full_name = self._get_role_full_name(
                            role_owner, role_name)

                    metadata = {
                        'vcs_url': role_vcs_url,
                        'role_format_version':
                            role_metadata['role_format_version']
                            if 'role_format_version' in role_metadata
                            else role_dir_info['role_format_version'],
                        'role_owner': role_owner,
                        'role_name': role_name,
                    }

                    metadata.update(
                        self._get_vcs_info(os.path.join(
                            role_path, role_dir_name))
                    )

                    if 'meta' in role_metadata:
                        metadata.update(
                            self._get_normalized_meta_main(
                                role_metadata['meta'])
                        )
                    if 'ansigenome' in role_metadata:
                        metadata.update(
                            self._get_normalized_meta_ansigenome(
                                role_metadata['ansigenome'])
                        )
                    if 'maintainer_nicks' in role_metadata:
                        nicks = role_metadata['maintainer_nicks']
                        metadata.setdefault('authors', [])

                        author_present = set([])
                        for author_item in metadata['authors']:
                            if author_item['nick'] in nicks:
                                author_present.add(author_item['nick'])
                                author_item['maintainer'] = True
                        if not author_present:
                            if self._strict:
                                raise Exception(
                                    "Nick(s) {nicks} are maintainers but"
                                    " no other meta information for them"
                                    " could be found in the repository."
                                    " Affected role: {role_full_name}".format(
                                        role_full_name=role_full_name,
                                        nicks=set(nicks).difference(
                                            author_present),
                                    )
                                )

                else:
                    # Legacy stuff.
                    role_name = role_dir_info['role_name']
                    role_owner = (role_dir_info['role_owner']
                                  if role_dir_info['role_owner']
                                  else self._role_owner)
                    if not role_owner:
                        raise Exception("Default role owner"
                                        " not given but required.")

                    role_vcs_url = self._get_repo_url(
                        role_owner,
                        role_name,
                    )
                    role_full_name = self._get_role_full_name(
                        role_owner,
                        role_name,
                    )

                    metadata = {
                        'vcs_url': role_vcs_url,
                        'role_format_version': '0.1.0',
                        'role_owner': role_owner,
                        'role_name': role_name,
                    }

                self._metadata.setdefault(role_full_name, {})
                self._metadata[role_full_name].update(metadata)

    def get_metadata(self, metadata=None):
        """
        Return public metadata.
        """

        if not metadata:
            metadata = self._metadata

        return metadata

        if not isinstance(metadata, dict):
            return metadata

        public_metadata = {}

        for k, v in metadata.items():
            if not k.startswith('_'):
                public_metadata[k] = self.get_metadata(v)

        return public_metadata

    def write_api_dir(self, api_dir):
        """
        Write metadata to API directory.
        The write is done to a temp directory which is later renamed to archive
        a atomic operation and to ensure that the API returns consistent data.
        """

        api_work_root_dir = api_dir
        api_dir = os.path.join(api_dir, 'v0')
        api_work_dir = api_dir + '_new'

        try:
            shutil.rmtree(api_work_dir)
        except OSError:
            pass

        os.makedirs(os.path.join(api_work_dir, 'role'))
        os.makedirs(os.path.join(api_work_dir, 'roles'))

        # API: /
        with open(os.path.join(api_work_dir, 'version'), 'w') as outfile:
            outfile.write('{}\n'.format(__version__))

        debops_api_base_dir = os.path.join(os.path.dirname(
            os.path.realpath(__file__)), '..')
        with open(os.path.join(api_work_dir, 'license'), 'w') as outfile:
            with open(os.path.join(debops_api_base_dir,
                                   'COPYRIGHT'), 'r') as copyright:
                outfile.write(copyright.read())
            outfile.write('\n\n\n')
            with open(os.path.join(debops_api_base_dir,
                                   'LICENSE'), 'r') as license:
                outfile.write(license.read())

        with open(os.path.join(debops_api_base_dir,
                               'README.rst'), 'r') as license:
            readme_html_string = reSTify(
                    license.read()).decode('utf-8', 'strict')
            with open(os.path.join(api_work_root_dir,
                                   'README.html'), 'w') as outfile:
                outfile.write(readme_html_string)
            with open(os.path.join(api_work_dir,
                                   'README.html'), 'w') as outfile:
                outfile.write(readme_html_string)

        # API: /role/
        for role_full_name, metadata in self.get_metadata().items():
            role_api_file = os.path.join(
                api_work_dir,
                'role',
                role_full_name + '.json'
            )
            with open(role_api_file, 'w') as outfile:
                json.dump(metadata, outfile, sort_keys=True)
                outfile.write('\n')

        # API: /roles/
        with open(os.path.join(api_work_dir,
                               'roles', 'count'), 'w') as outfile:
            outfile.write('{}\n'.format(len(self.get_metadata().keys())))

        metadata_per_owner = {}
        for role_full_name, metadata in self.get_metadata().items():
            role_owner = metadata['role_owner']
            metadata_per_owner.setdefault(role_owner, {})
            metadata_per_owner[role_owner][role_full_name] = metadata

        for role_owner, metadata in metadata_per_owner.items():
            with open(os.path.join(api_work_dir, 'roles',
                                   role_owner + '.list'),
                      'w') as outfile:
                role_list = self.get_metadata(metadata).keys()
                outfile.write('\n'.join(sorted(role_list)) + '\n')
            with open(os.path.join(api_work_dir,
                                   'roles', role_owner + '.json'),
                      'w') as outfile:
                json.dump(self.get_metadata(metadata), outfile, sort_keys=True)
                outfile.write('\n')
            with open(os.path.join(api_work_dir,
                                   'roles', 'count:' + role_owner),
                      'w') as outfile:
                outfile.write('{}\n'.format(
                    len(self.get_metadata(metadata).keys())))

        try:
            shutil.rmtree(api_dir)
        except OSError:
            pass

        os.rename(api_work_dir, api_dir)


if __name__ == '__main__':
    from argparse import ArgumentParser

    args_parser = ArgumentParser(
        description=__doc__,
    )
    args_parser.add_argument(
        '-V', '--version',
        action='version',
        version='%(prog)s {version}'.format(version=__version__)
    )
    args_parser.add_argument(
        '-d', '--debug',
        help="Print lots of debugging statements.",
        action='store_const',
        dest='loglevel',
        const=logging.DEBUG,
        default=logging.WARNING,
    )
    args_parser.add_argument(
        '-v', '--verbose',
        help="Be verbose.",
        action='store_const',
        dest='loglevel',
        const=logging.INFO,
    )
    args_parser.add_argument(
        '-n', '--no-strict',
        help="Do not exit immediately when there is a inconsistency.",
        dest='strict',
        action='store_false',
        default=True,
    )
    args_parser.add_argument(
        '-i', '--github-api-response-file',
        help="Responds file from a /users/:username/repos"
        " or /orgs/:org/repos API query.",
    )
    args_parser.add_argument(
        '-r', '--role-path',
        help="Base directory where all roles are available.",
    )
    args_parser.add_argument(
        '-o', '--role-owner',
        help="Default role owner if not available from Ansible role metadata.",
        default='debops',
    )
    args_parser.add_argument(
        '-a', '--api-dir',
        help="Write the static parts of api.debops.org"
        " to the given directory."
        " Note that all files in this directory are going to be"
        " overwritten or deleted.",
    )
    args_parser.add_argument(
        '-D', '--docs-url-pattern',
        help="Documentation URL for each role.",
        default='https://docs.debops.org/en/latest/'
        'ansible/roles/ansible-{role_name}/docs/index.html',
    )
    args_parser.add_argument(
        '-C', '--changelog-url-pattern',
        help="Changelog URL for each role.",
        default='https://docs.debops.org/en/latest/'
        'ansible/roles/ansible-{role_name}/docs/changelog.html',
    )
    args_parser.add_argument(
        '-t', '--test-mode',
        help="Make the output reproducible by normalizing changing peaces like"
        "timestamps.",
        action='store_true',
        default=False,
    )
    args = args_parser.parse_args()
    if not args.github_api_response_file and not args.role_path:
        args_parser.print_help()
        sys.exit(1)

    logger = logging.getLogger(__file__)
    logging.basicConfig(
        format='%(levelname)s: %(message)s',
        level=args.loglevel,
    )

    debops_metadata = DebOpsAPI(
        strict=args.strict,
        docs_url_pattern=args.docs_url_pattern,
        changelog_url_pattern=args.changelog_url_pattern,
        role_owner=args.role_owner,
        test_mode=args.test_mode,
    )

    if args.github_api_response_file:
        debops_metadata.read_github_repos_api_file(
            args.github_api_response_file,
        )

    if args.role_path:
        debops_metadata.read_role_metadata(args.role_path)

    debops_metadata.gen_role_metadata()

    if args.api_dir:
        debops_metadata.write_api_dir(args.api_dir)

    logging.info("Metadata:\n{}".format(
        pprint.pformat(debops_metadata.get_metadata()),
    ))



from tests.icinga_helper import IcingaHelper

testinfra_hosts = ['ansible://bind-host']

IcingaHelper.icinga_host = 'bind-host'


class TestChecks(IcingaHelper):

    def test_host(self):
        assert 'postfix-host' in self.get_hosts(host='postfix-host')

    def test_service(self, host):
        #  r = self.get_client().objects.list('Service', joins=['host.name'])
        with host.sudo():
            host.run("systemctl restart icinga2")
        assert self.is_service_ok('postfix-host!Check smtps TLS certificate')



from tests.icinga_helper import IcingaHelper

testinfra_hosts = ['ansible://icinga-host']


class TestChecks(IcingaHelper):

    def test_icinga_host(self):
        hosts = self.get_hosts()
        assert 'bind-host' in hosts
        assert 'deleted-host' not in hosts

    def test_icinga_service(self, host):
        #  r = self.get_client().objects.list('Service', joins=['host.name'])
        with host.sudo():
            host.run("systemctl restart icinga2")
        assert self.is_service_ok('bind-host!ping4')
        domain = host.run("hostname -d").stdout.strip()
        assert self.is_service_ok('bind-host!Zone test.' + domain)
        assert self.is_service_ok('bind-host!Zone ' + domain)



import logging
import os
import pytest
import sh
import time

from enough import settings
from enough.common import tcp
from enough.common.retry import retry
from enough.common.openstack import OpenStack

from tests import prepare_config_dir


@pytest.fixture(autouse=True, scope='session')
def debug_enough():
    logging.getLogger('enough').setLevel(logging.DEBUG)


@pytest.fixture
def tcp_port():
    return str(tcp.free_port())


def docker_conf():
    if 'DOCKER_HOST' in os.environ:
        return {
            'namespace': 'myaccount',
        }
    else:
        return {
            'port': tcp.free_port(),
        }


@pytest.fixture
def docker_args():
    return docker_conf()


@pytest.fixture
def docker_options():
    options = []
    for k, v in docker_conf().items():
        k = k.replace('_', '-')
        options.append('--' + k)
        options.append(v)
    return options


class DockerLeftovers(Exception):
    pass


@retry(DockerLeftovers, tries=7)
def docker_cleanup(prefix):
    leftovers = []
    for container in sh.docker.ps('--all', '--format', '{{ .Names }}', _iter=True):
        container = container.strip()
        if prefix in container:
            sh.docker.rm('-f', container, _ok_code=[0, 1])
            leftovers.append('container(' + container + ')')
    for network in sh.docker.network.ls('--format', '{{ .Name }}', _iter=True):
        network = network.strip()
        if prefix in network:
            sh.docker.network.rm(network, _ok_code=[0, 1])
            leftovers.append('network(' + network + ')')
    for image in sh.docker.images('--format', '{{ .Repository }}:{{ .Tag }}', _iter=True):
        image = image.strip()
        if image.startswith(prefix):
            sh.docker.rmi('--no-prune', image, _ok_code=[0, 1])
            leftovers.append('image(' + image + ')')
    if leftovers:
        raise DockerLeftovers('scheduled removal of ' + ' '.join(leftovers))


@pytest.fixture
def docker_name():
    prefix = 'enough_test_' + str(int(time.time()))
    yield prefix
    docker_cleanup(prefix)


@pytest.fixture
def openstack_name():
    prefix = 'enough_test_' + str(int(time.time()))
    yield prefix
    o = OpenStack('.')
    o.destroy_everything(prefix)


@pytest.fixture
def tmp_config_dir(monkeypatch, tmpdir):
    enough_dot_dir = str(tmpdir)
    domain = 'enough.community'
    monkeypatch.setenv('ENOUGH_DOT', enough_dot_dir)
    monkeypatch.setenv('ENOUGH_DOMAIN', domain)
    config_dir = prepare_config_dir(domain, enough_dot_dir)
    monkeypatch.setattr(settings, 'CONFIG_DIR', config_dir)


# Some tests mock os.path.exists using the mocker fixture, use trylast in
# order to avoid any interference
@pytest.hookimpl(trylast=True)
def pytest_runtest_teardown(item, nextitem):
    assert os.path.isdir('.git')  # check cwd is the source repository
    assert os.path.isdir('inventory/group_vars/all')  # with the expected structure

    check_generated_files = [
        'inventory/group_vars/all/private-key.yml',
    ]
    if os.environ.get('GITLAB_CI'):
        # contributors are allowed to define such files, CI isn't
        check_generated_files.extend((
            'infrastructure_key',
            'infrastructure_key.pub',
        ))
    # first remove all these files in order to not interfere with the next
    # tests then fail
    found = []
    for path in check_generated_files:
        if os.path.exists(path):
            found.append(path)
            os.remove(path)
    assert not found



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) 2014, Ilya Barsukov <barsukov@selectel.ru>, Selectel LLC
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

# import module snippets
from ansible.module_utils.basic import *
import os


DOCUMENTATION = """
---
module: btrfs_subvolume
short_description: Provides `create` and `delete` subvolumes methods.
description:
     - The M(btrfs_subvolume) module takes the command name followed by
       a list of space-delimited arguments.
     - The given command will be executed on all selected nodes.
version_added: 0.1
author: Ilya Barsukov
options:
    state:
        description:
            - creates or deletes subvolume
        required: true
        choices: ["present", "absent"]
    path:
        description:
            - subvolume absolute path
        required: true
        default: null
    qgroups:
        required: false
        description:
            - list of qgroup ids, adds the newly created subvolume to a qgroup
        default: []
    commit:
        required: false
        description:
            - wait for transaction commit at the end of the operation
              or of the each
        choices: ["each", "after", "no"]
        default: "no"
    recursive:
        required: false
        choices: [true, false]
        description:
            - create or delete subvolumes recursively
        default: false
"""

EXAMPLES = """
# Example for Ansible Playbooks.
- name: Recursive create given subvolume path
  btrfs_subvolume:
    state: 'present'
    path: '/storage/test/test1/test2'
    qgroups:
      - '1/100'
      - '1/101'
    recursive: True

- name: Delete given Btrfs subvolume
  btrfs_subvolume:
    state: 'absent'
    path: '/storage/test/test1/test2'
    commit: 'each'
"""


def get_subvolumes(path, subs=None):
    if subs is None:
        subs = []
    subvolumes = os.listdir(path)
    for sub in subvolumes:
        sub = os.path.sep.join([path, sub])

        if not os.path.isdir(sub):
            continue

        subs.append(sub)
        subs = get_subvolumes(sub, subs=subs)

    return subs


def main():
    module = AnsibleModule(
        argument_spec=dict(
            state=dict(required=True, choices=['present', 'absent'],
                       type='str'),
            path=dict(required=True, default=None, type='str'),
            qgroups=dict(default=[], type='list'),
            commit=dict(default='no', choices=['after', 'each', 'no'],
                        type='str'),
            recursive=dict(default='False', type='bool')
        ),
        supports_check_mode=True
    )
    result = {
        'changed': False,
        'commands': [],
        'check': module.check_mode
    }
    param_path = module.params['path'].rstrip(os.path.sep)

    if module.params['state'] == 'present':
        # Creating subvolume
        if not os.path.exists(param_path) and not module.params['recursive']:
            cmd = 'btrfs subvolume create {qgroups} {subvolume}'.format(
                qgroups=' -i '.join(['']+module.params['qgroups']),
                subvolume=param_path,
            )
            result['commands'].append(cmd)
            if not module.check_mode:
                module.run_command(cmd, check_rc=True)
                result['changed'] = True

        elif module.params['recursive']:
            # Check parent subvolumes and create it if they doesnt exist
            parents = param_path.split(os.path.sep)
            for idx, subvolume in enumerate(parents):
                if len(subvolume) == 0:
                    continue

                subvolume = os.path.sep.join(parents[:idx+1])

                if not os.path.exists(subvolume):
                    cmd = ('btrfs subvolume create '
                           '{qgroups} {subvolume}').format(
                        qgroups=' -i '.join(['']+module.params['qgroups']),
                        subvolume=subvolume,
                    )

                    result['commands'].append(cmd)
                    if not module.check_mode:
                        module.run_command(cmd, check_rc=True)
                        result['changed'] = True

    elif module.params['state'] == 'absent':
        # Delete subvolume
        commit = ''
        if module.params['commit'] != 'no':
            commit = '--commit-{}'.format(module.params['commit'])

        if os.path.exists(param_path):
            if not module.params['recursive']:
                cmd = 'btrfs subvolume delete {commit} {subvolume}'.format(
                    commit=commit, subvolume=param_path
                )
                result['commands'].append(cmd)
                if not module.check_mode:
                    module.run_command(cmd, check_rc=True)
                    result['changed'] = True

            elif module.params['recursive']:
                # reversed parent directories from end to beginning
                subvolumes = get_subvolumes(param_path)
                subvolumes.insert(0, param_path)

                for sub in reversed(subvolumes):

                    if os.path.exists(sub):
                        cmd = ('btrfs subvolume delete '
                               '{commit} {subvolume}').format(
                            commit=commit, subvolume=sub
                        )

                        result['commands'].append(cmd)
                        if not module.check_mode:
                            module.run_command(cmd, check_rc=True)
                            result['changed'] = True

    if module.check_mode and result['commands']:
        result['changed'] = True

    if not module.check_mode:
        del result['check']

    if not result['commands']:
        del result['commands']

    module.exit_json(**result)


if __name__ == '__main__':
    main()



import ast
import base64
import datetime
import json
import logging
import os
import re
import requests
import sh
import socket
import textwrap
import time
import yaml

from enough import settings
from enough.common import ansible_utils
from enough.common.retry import retry
from enough.common.dotenough import Hosts

log = logging.getLogger(__name__)


class OpenStackBase(object):

    INTERNAL_NETWORK = 'internal'
    INTERNAL_NETWORK_CIDR = '10.20.30.0/24'

    def __init__(self, config_dir, **kwargs):
        self.args = kwargs
        self.config_dir = config_dir
        self.config_file = f'{self.config_dir}/inventory/group_vars/all/clouds.yml'
        self.o = sh.openstack.bake(
            '--os-cloud', kwargs.get('cloud', 'production'),
            _tee=True,
            _tty_out=False,
            _out=lambda x: log.info(x.strip()),
            _err=lambda x: log.info(x.strip()),
            _env={'OS_CLIENT_CONFIG_FILE': self.config_file},
        )

    @staticmethod
    @retry((socket.timeout, ConnectionRefusedError, OSError), 9)
    def wait_for_ssh(ip, port):
        log.info('Check if SSH is available on %s:%s', ip, port)
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(5)
        s.connect((ip, int(port)))
        line = s.makefile("rb").readline()
        assert line.startswith(b'SSH-')


class Stack(OpenStackBase):

    class IPNotFound(Exception):
        pass

    def __init__(self, config_dir, definition=None, **kwargs):
        super().__init__(config_dir, **kwargs)
        log.info(f'OS_CLIENT_CONFIG_FILE={self.config_file}')
        self.definition = definition

    def get_template(self):
        return f'{settings.SHARE_DIR}/playbooks/infrastructure/template-host.yaml'

    def set_public_key(self, path):
        self.public_key = open(path).read().strip()

    def _create_or_update(self, action):
        d = self.definition
        parameters = [
            f'--parameter=public_key={self.public_key}',
        ]
        if 'flavor' in d:
            parameters.append(f'--parameter=flavor={d["flavor"]}')
        if 'image' in d:
            parameters.append(f'--parameter=image={d["image"]}')
        if 'port' in d:
            parameters.append(f'--parameter=port={d["port"]}')
        if 'volumes' in d and int(d['volumes'][0]['size']) > 0:
            parameters.append(f"--parameter=volume_size={d['volumes'][0]['size']}")
            parameters.append(f"--parameter=volume_name={d['volumes'][0]['name']}")
        if d.get('network'):
            parameters.append(f'--parameter=network={d["network"]}')
            parameters.append(
                f'--parameter=network_interface_unconfigured={d["network_interface_unconfigured"]}')
            parameters.append(
                f'--parameter=network_interface_routed={d["network_interface_routed"]}')
            parameters.append(
                f'--parameter=network_interface_not_routed={d["network_interface_not_routed"]}')
        log.info(f'_create_or_update: {d["name"]} {parameters}')
        self.o.stack(action, d['name'],
                     '--wait', '--timeout=600',
                     '--template', self.get_template(),
                     *parameters)

    def list(self):
        return [
            s.strip() for s in
            self.o.stack.list('--format=value', '-c', 'Stack Name', _iter=True)
        ]

    def create_or_update(self, return_on_create=False):
        if self.definition['name'] not in self.list():
            self.create()
            if return_on_create:
                return None
        else:
            self._create_or_update('update')
        self.create_internal_network()
        self.connect_internal_network()
        info = {
            'ipv4': self.get_ipv4(),
            'port': self.definition['port'],
        }
        if not self.internal_only() or self.args.get('route_to_internal', True):
            self.wait_for_ssh(info['ipv4'], info['port'])
        Hosts(self.config_dir).create_or_update(
            self.definition['name'], info['ipv4'], info['port'])
        return info

    @retry((IPNotFound, sh.ErrorReturnCode_1), 5)
    def get_ipv4(self):
        server = self.definition['name']
        o = OpenStack(self.config_dir, **self.args)
        if self.internal_only():
            ipv4 = o.server_ip_in_network(
                server, self.definition.get('internal_network', OpenStackBase.INTERNAL_NETWORK))
        else:
            ipv4 = o.server_ip_in_network(
                server, self.definition.get('network', 'Ext-Net'))
        if ipv4 is None:
            raise Stack.IPNotFound()
        return ipv4

    def create_internal_network(self):
        network = self.definition.get('internal_network', OpenStackBase.INTERNAL_NETWORK)
        cidr = self.definition.get('internal_network_cidr', OpenStackBase.INTERNAL_NETWORK_CIDR)
        o = OpenStack(self.config_dir, **self.args)
        o.network_and_subnet_create(network, cidr)

    def connect_internal_network(self):
        server = self.definition['name']
        network = self.definition.get('internal_network', OpenStackBase.INTERNAL_NETWORK)
        o = OpenStack(self.config_dir, **self.args)
        if o.server_connected_to_network(server, network):
            return
        self.o.server.add.network(server, network)

    def internal_only(self):
        return self.definition.get('network_internal_only', False)

    def create(self):
        try:
            log.info('create or update')
            self._create_or_update('create')
        except sh.ErrorReturnCode_1:
            log.info('retry create or update')
            # retry once because there is no way to increase the timeout and create fails often
            self._create_or_update('update')

    def delete_no_wait(self):
        name = self.definition['name']
        if name not in self.list():
            Hosts(self.config_dir).delete(name)
            return False

        self.o.stack.delete('--yes', name)
        return True

    def delete_wait(self):
        name = self.definition['name']

        @retry(AssertionError, 9)
        def wait_is_deleted():
            assert name not in self.list(), f'{name} deletion in progress'
        wait_is_deleted()

        Hosts(self.config_dir).delete(name)

        network = self.definition.get('internal_network', OpenStackBase.INTERNAL_NETWORK)
        o = OpenStack(self.config_dir, **self.args)
        o.delete_security_group(name)
        o.network_delete_if_not_used(network)

    def delete(self):
        if self.delete_no_wait():
            self.delete_wait()


class Heat(OpenStackBase):

    def get_stack_definitions(self, share_dir=settings.SHARE_DIR):
        ansible = ansible_utils.Ansible(self.config_dir, share_dir,
                                        self.args.get('inventory'))
        return ansible.get_hostvars(variable=None)

    def get_stack_definition(self, host):
        hostvars = self.get_stack_definitions()[host]
        return Heat.hostvars_to_stack(host, hostvars)

    @staticmethod
    def hostvars_to_stack(host, hv):
        if hv.get('network_internal_only'):
            network_interface_unconfigured = hv['network_primary_interface']
            network_interface_routed = hv['network_secondary_interface']
            network_interface_not_routed = 'noname'
        else:
            network_interface_unconfigured = 'noname'
            network_interface_routed = hv['network_primary_interface']
            network_interface_not_routed = hv['network_secondary_interface']
        definition = {
            'name': host,
            'port': hv.get('ansible_port', '22'),
            'flavor': hv['openstack_flavor'],
            'image': hv['openstack_image'],
            'network': hv.get('openstack_network'),
            'network_internal_only': hv.get('network_internal_only'),
            'network_interface_unconfigured': network_interface_unconfigured,
            'network_interface_routed': network_interface_routed,
            'network_interface_not_routed': network_interface_not_routed,
            'internal_network': OpenStackBase.INTERNAL_NETWORK,
            'internal_network_cidr': hv.get('openstack_internal_network_cidr',
                                            OpenStackBase.INTERNAL_NETWORK_CIDR),
        }
        if 'openstack_volumes' in hv:
            definition['volumes'] = hv['openstack_volumes']
        return definition

    def host_from_volume(self, name):
        for host, definition in self.get_stack_definitions().items():
            for v in definition.get('openstack_volumes', []):
                if v['name'] == name:
                    return host
        return None

    def is_working(self):
        # retry to verify the API is stable
        for _ in range(5):
            try:
                self.o.stack.list()
            except sh.ErrorReturnCode_1:
                return False
        return True

    def create_missings(self, names, public_key):
        return self.create_or_update(Hosts(self.config_dir).missings(names), public_key)

    def create_or_update(self, names, public_key):
        r = {}
        created = []
        #
        # Launch the creation of all stacks in // and do not wait for them to complete
        #
        for name in names:
            s = Stack(self.config_dir, self.get_stack_definition(name),
                      **self.args)
            s.set_public_key(public_key)
            info = s.create_or_update(return_on_create=True)
            if info:
                r[name] = info
            else:
                created.append(name)
        #
        # Verify all stacks previously launched complete as expected
        #
        for name in created:
            s = Stack(self.config_dir, self.get_stack_definition(name),
                      **self.args)
            s.set_public_key(public_key)
            info = s.create_or_update(return_on_create=True)
            r[name] = s.create_or_update()
        return r

    def create_test_subdomain(self, domain):
        d = f"{self.config_dir}/inventory/group_vars/all"
        assert os.path.exists(d)
        if 'bind-host' not in Stack(self.config_dir, **self.args).list():
            return None
        h = Heat(self.config_dir, **self.args)
        s = Stack(self.config_dir, h.get_stack_definition('bind-host'),
                  **self.args)
        s.set_public_key(f'{self.config_dir}/infrastructure_key.pub')
        bind_host = s.create_or_update()

        # reverse so the leftmost part varies, for human readability
        s = str(int(time.time()))[::-1]
        subdomain = base64.b32encode(s.encode('ascii')).decode('ascii').lower()

        fqdn = f'{subdomain}.test.{domain}'
        log.info(f'creating test subdomain {fqdn}')

        token = os.environ['ENOUGH_API_TOKEN']

        r = requests.post(f'https://api.{domain}/delegate-test-dns/',
                          headers={'Authorization': f'Token {token}'},
                          json={
                              'name': subdomain,
                              'ip': bind_host['ipv4'],
                          })
        r.raise_for_status()
        open(f'{d}/domain.yml', 'w').write(textwrap.dedent(f"""\
        ---
        domain: {fqdn}
        """))
        return fqdn


class OpenStackLeftovers(Exception):
    pass


class OpenStackBackupCreate(Exception):
    pass


class OpenStackShutoff(Exception):
    pass


class OpenStackVolumeResizeMissing(Exception):
    pass


class OpenStackVolumeResizeMismatch(Exception):
    pass


class OpenStackVolumeResizeShrink(Exception):
    pass


class OpenStack(OpenStackBase):

    def __init__(self, config_dir, **kwargs):
        super().__init__(config_dir, **kwargs)
        self.config = yaml.safe_load(open(self.config_file))

    @retry(OpenStackLeftovers, tries=7)
    def destroy_everything(self, prefix):
        log.info('destroy_everything acting')
        leftovers = []

        def delete_snapshots():
            snapshots = self.o.volume.snapshot.list('--format=json',
                                                    '-c', 'Name', '-c', 'ID')
            for snapshot in json.loads(snapshots.stdout):
                if prefix is None or prefix in snapshot['Name']:
                    leftovers.append(f'snapshot({snapshot["ID"]})')
                    self.o.volume.snapshot.delete(snapshot['ID'])

        def delete_stacks():
            r = self.o.stack.list('--format=json', '-c', 'Stack Name', '-c', 'Stack Status')
            for name, status in [(x["Stack Name"], x["Stack Status"])
                                 for x in json.loads(r.stdout)]:
                if prefix is None or prefix in name:
                    leftovers.append(f'stack({name})')
                    if status == 'DELETE_FAILED' or not status.startswith('DELETE'):
                        self.o.stack.delete('--yes', '--wait', name)

        def delete_images():
            for image in self.o.image.list('--private', '--format=value', '-c', 'Name', _iter=True):
                image = image.strip()
                if prefix is None or prefix in image:
                    leftovers.append(f'image({image})')
                    self.o.image.delete(image)

        def delete_volumes():
            volumes = self.o.volume.list('--format=json', '-c', 'Name',
                                         '-c', 'ID')

            for volume in json.loads(volumes.stdout):
                if prefix is None or prefix in volume['Name']:
                    leftovers.append(f'volume({volume["ID"]})')
                    self.o.volume.delete(volume['ID'])

        def delete_networks():
            for network in self.o.network.list('--no-share', '--format=value', '-c', 'Name',
                                               _iter=True):
                network = network.strip()
                if prefix is None or prefix in network:
                    leftovers.append(f'network({network})')
                    self.o.network.delete(network)

        #
        # There may be complex interdependencies between resources and
        # no easy way to figure them out. For instance, say there
        # exists a volume created from a snapshot of a volume created
        # by a stack. The stack cannot be deleted befor the volume created
        # from the snapshot is deleted. Because the snapshot cannot be deleted
        # before all volumes created from it are deleted. And the volumes from
        # which the snapshot are created cannot be deleted before all their
        # snapshots are deleted.
        #
        for f in (delete_snapshots, delete_stacks, delete_images, delete_volumes, delete_networks):
            try:
                f()
            except sh.ErrorReturnCode_1:
                pass

        if leftovers:
            raise OpenStackLeftovers('scheduled removal of ' + ' '.join(leftovers))

    def run(self, *args):
        return self.o(*args)

    def delete_security_group(self, name):
        try:
            self.o.security.group.delete(name)
        except sh.ErrorReturnCode_1:
            return False
        return True

    def network_exists(self, name):
        network = self.o.network.list('--format=value', '-c', 'Name', '--name', name)
        return network.strip() == name

    def network_create(self, name):
        if not self.network_exists(name):
            self.o.network.create(name)

    def network_delete_if_not_used(self, name):
        if not self.network_exists(name):
            return

        ports = self.o.port.list('--network', name, '--format=json', '-c',
                                 'ID', '-c', 'device_owner')
        ports = [port for port in json.loads(str(ports))
                 if port['device_owner'].startswith('compute:')]
        if not ports:
            self.o.network.delete(name)

    def subnet_exists(self, name):
        subnet = self.o.subnet.list('--format=value', '-c', 'Name', '--name', name)
        return subnet.strip() == name

    def subnet_create(self, name, cidr):
        if not self.subnet_exists(name):
            self.o.subnet.create('--subnet-range', cidr,
                                 '--dns-nameserver=8.8.8.8',
                                 '--network', name,
                                 name)

    def subnet_update_dns(self, name, *dns):
        args = [f'--dns-nameserver={ip}' for ip in dns]
        args.append(name)
        self.o.subnet.set('--no-dns-nameserver', *args)

    def network_and_subnet_create(self, name, cidr):
        self.network_create(name)
        self.subnet_create(name, cidr)

    def server_connected_to_network(self, server, network):
        return self.server_ip_in_network(server, network)

    def server_port_list(self, server, network):
        return self.o.port.list('--server', server, '--network', network,
                                '--format=value', '-c', 'Fixed IP Addresses')

    def server_ip_in_network(self, server, network):
        info = self.server_port_list(server, network).strip()
        if info == '':
            return None
        pattern = r"^[0-9.]+$"
        found = None
        for subnet in ast.literal_eval(info):
            if re.match(pattern, subnet['ip_address']):
                found = subnet['ip_address']
        return found

    def backup_date(self):
        return datetime.datetime.today().strftime('%Y-%m-%d')

    def backup_create(self, volumes):
        if len(volumes) == 0:
            volumes = [x.strip() for x in self.o.volume.list('--format=value', '-c', 'Name')]
        date = self.backup_date()
        snapshots = self._backup_map()
        count = 0
        for volume in volumes:
            s = f'{date}-{volume}'
            if s not in snapshots:
                self.o.volume.snapshot.create('--force', '--volume', volume, s)
                count += 1
        self._backup_available(volumes, date)
        return count

    def _backup_map(self):
        return dict(self._backup_list())

    def _backup_list(self):
        r = self.o.volume.snapshot.list('--format=json', '-c', 'Name', '-c', 'Status',
                                        '--limit', '5000')
        return [(x["Name"], x["Status"]) for x in json.loads(r.stdout)]

    @retry(OpenStackBackupCreate, tries=7)
    def _backup_available(self, volumes, date):
        available = []
        waiting = []
        for name, status in self._backup_list():
            if not name.startswith(date):
                continue
            if status == "available":
                available.append(name)
            else:
                waiting.append(f'{status} {name}')
        available = ",".join(available)
        waiting = ",".join(waiting)
        progress = f'WAITING on {waiting}\nAVAILABLE {available}'
        log.debug(progress)
        if len(waiting) > 0:
            raise OpenStackBackupCreate(progress)

    def backup_prune(self, days):
        before = (datetime.datetime.today() - datetime.timedelta(days)).strftime('%Y-%m-%d')
        count = 0
        for name, status in self._backup_list():
            if name[:10] > before:
                continue
            self.o.volume.snapshot.delete(name)
            count += 1
        return count

    def volume_prune(self, days):
        r = self.o.volume.snapshot.list(
            '--format=json', '--long', '-c', 'Volume', '--limit', '5000')
        volumes_with_snapshots = set([x["Volume"] for x in json.loads(r.stdout)])
        r = self.o.volume.list('--format=json', '-c', 'Name', '-c', 'ID', '--limit', '5000')
        before = (datetime.datetime.today() - datetime.timedelta(days)).strftime('%Y-%m-%d')
        info = {
            'no_date_prefix': [],
            'has_snapshots': [],
            'deleted': [],
            'recent': [],
        }
        for volume, volume_id in set([(x["Name"], x["ID"]) for x in json.loads(r.stdout)]):
            if re.match(r'^\d\d\d\d\-\d\d-\d\d-', volume):
                if volume_id in volumes_with_snapshots:
                    info['has_snapshots'].append(volume)
                elif volume[:10] < before:
                    self.o.volume.delete(volume)
                    info['deleted'].append(volume)
                else:
                    info['recent'].append(volume)
            else:
                info['no_date_prefix'].append(volume)
        return info

    @retry(OpenStackShutoff, tries=5)
    def server_wait_shutoff(self, host):
        status = self.o.server.show('--format=value', '-c=status', host)
        status = status.strip()
        if status != 'SHUTOFF':
            raise OpenStackShutoff('Unexpected status: %r' % status)

    def replace_volume(self, host, volume, delete_volume):
        self.o.server.stop(host)
        self.server_wait_shutoff(host)
        attached = self.o.server.show(
            '--format=yaml', '-c', 'volumes_attached', host).stdout.strip()
        current_volume_id = yaml.safe_load(attached)['volumes_attached'][0]['id']
        current_volume = self.o.volume.show(
            '--format=value', '-c', 'name', current_volume_id).strip()
        self.o.server.remove.volume(host, current_volume_id)
        # fetch volume id now because next block might assign an existing volume
        # name to another volume
        volume_id = self.o.volume.show('--format=value', '-c', 'id', volume).stdout.strip()
        if delete_volume:
            # using current_volume_id here is not possible unfortunately
            self.o.volume.delete(current_volume)
        else:
            date = self.backup_date()
            self.o.volume.set('--name', f'{date}-{current_volume}', current_volume_id)
        self.o.volume.set('--name', current_volume, volume_id)
        self.o.server.add.volume(host, volume_id)
        self.o.server.start(host)

    def volume_resize(self, host, volume, size):
        attached = self.o.server.show(
            '--format=yaml', '-c', 'volumes_attached', host).stdout.strip()
        volumes_attached = yaml.safe_load(attached)['volumes_attached']
        if not volumes_attached:
            raise OpenStackVolumeResizeMissing(
                f'{host} is not attached {volume}, it is not attached to any volume')
        current_volume_id = volumes_attached[0]['id']
        current_volume = self.o.volume.show(
            '--format=value', '-c', 'name', current_volume_id).strip()
        if current_volume != volume:
            raise OpenStackVolumeResizeMismatch(
                f'{host} is not attached {volume}, it is attached to {current_volume}')
        current_size = int(self.o.volume.show('-c', 'size', '--format=value', volume).strip())
        if current_size == size:
            log.info(f'resize of {host} volume {volume} is not necessary, it already is {size}GB')
            return False
        if current_size > size:
            raise OpenStackVolumeResizeShrink(
                f'{host} {volume} is {current_size}GB and cannot be shrinked to {size}GB')
        self.o.server.stop(host)
        self.server_wait_shutoff(host)
        self.o.server.remove.volume(host, volume)
        self.o.volume.set('--size', size, volume)
        self.o.server.add.volume(host, volume)
        self.o.server.start(host)
        return True

    def region_empty(self):
        volumes = self.o.volume.list()
        servers = self.o.server.list()
        images = self.o.image.list('--private')
        return volumes.strip() == '' and servers.strip() == '' and images.strip() == ''



testinfra_hosts = ['ansible://internal-host']


def test_vpn_route(host):
    cmd = host.run("hostname")
    print(cmd.stdout)
    print(cmd.stderr)
    assert cmd.rc == 0
    assert 'internal-host' in cmd.stdout



import pytest
import requests
from tests.infrastructure import get_driver
import yaml

testinfra_hosts = ['ansible://bind-host']


def get_address(inventory):
    vars_dir = f'{inventory}/group_vars/all'
    return 'icinga.' + yaml.safe_load(
        open(vars_dir + '/domain.yml'))['domain']


def test_icingaweb2_login_screen(host, pytestconfig):
    if get_driver() == 'docker':
        pytest.skip("no letsencrypt when running docker")
    address = get_address(pytestconfig.getoption("--ansible-inventory"))
    proto_srv = f"https://{address}"
    s = requests.Session()
    r = s.get(proto_srv+'/icingaweb2/authentication/login', timeout=20, verify='certs')
    cookies = dict(r.cookies)
    r = s.get(proto_srv+'/icingaweb2/authentication/login?_checkCookie=1',
              cookies=cookies, timeout=5, verify='certs')
    r.raise_for_status()
    assert 'Icinga Web 2 Login' in r.text



from django.conf import settings
import os
import requests
import time


#
# Not using python-gitlab because it does not support /oauth/token
# see https://github.com/python-gitlab/python-gitlab/issues/753
#
#
# https://docs.gitlab.com/ce/api/oauth2.html#resource-owner-password-credentials
#
class GitLab(object):

    def __init__(self, url):
        self.url = url
        self._session()

    def _session(self):
        self.s = requests.Session()
        if 'REQUESTS_CA_BUNDLE' not in os.environ:
            self.s.verify = settings.CERTS_DIR
        self.s.api = self.url + '/api/v4'

    def login(self, username, password):
        r = self.s.post(self.url + '/oauth/token', json={
            'username': username,
            'password': password,
            'grant_type': 'password',
        })
        r.raise_for_status()
        self.set_token(r.json()['access_token'])

    def set_token(self, token):
        self.s.headers['Authorization'] = f'Bearer {token}'

    def get_namespace(self, user):
        r = self.s.get(self.s.api + '/namespaces?search=' + user)
        r.raise_for_status()
        return r.json()[0]['id']

    def group_members(self, group):
        r = self.s.get(self.s.api + f'/groups/{group}/members')
        r.raise_for_status()
        return r.json()

    def is_member_of_group(self, group, username):
        return any([x['username'] == username for x in self.group_members(group)])

    def is_self_member_of_group(self, group):
        r = self.s.get(self.s.api + f'/user')
        r.raise_for_status()
        user = r.json()
        return self.is_member_of_group(group, user['username'])

    def recreate_project(self, user, project):
        namespace_id = self.get_namespace(user)
        r = self.s.get(self.s.api + '/projects/' + user + '%2F' + project)
        if r.status_code == requests.codes.ok:
            r = self.s.delete(self.s.api + '/projects/' + user + '%2F' + project)
            r.raise_for_status()
        for _ in range(10):
            r = self.s.post(self.s.api + '/projects', data={
                "name": project,
                "namespace_id": int(namespace_id),
                "visibility": "public",
            })
            time.sleep(5)
            if r.status_code == 201:
                break
            print(str(r.text))
        r.raise_for_status()

    def create_api_application(self, domain):
        callbacks = [
            f'https://api.{domain}/',
            f'https://api.{domain}/accounts/gitlab/',
            f'https://api.{domain}/accounts/gitlab/login/callback/',
        ]
        r = self.s.post(self.s.api + '/applications', json={
            'name': 'api',
            'redirect_uri': "\n".join(callbacks),
            'scopes': "api\nread_user",
            })
        r.raise_for_status()
        j = r.json()
        return j['application_id'], j['secret']

    def ensure_group_exists(self, name, **kwargs):
        r = self.s.get(f'{self.s.api}/groups/{name}')
        if r.status_code == 200:
            return
        args = {'name': name, 'path': name}
        args.update(kwargs)
        r = self.s.post(f'{self.s.api}/groups', json=args)
        r.raise_for_status()



import argparse
from cliff.command import Command

from enough import settings
from enough.common import ansible_utils


class Playbook(Command):
    "Playbook"

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument('args', nargs=argparse.REMAINDER)
        return parser

    def take_action(self, parsed_args):
        args = vars(self.app.options)
        args.update(vars(parsed_args))
        playbook = ansible_utils.Playbook(settings.CONFIG_DIR, settings.SHARE_DIR)
        playbook.run_from_cli(**args)



import json

from enough import cmd


def test_service(capsys, mocker, tmp_config_dir):
    # do not tamper with logging streams to avoid
    # ValueError: I/O operation on closed file.
    mocker.patch('cliff.app.App.configure_logging')
    r = {'fqdn': 'name.domain.com'}
    mocker.patch('enough.common.service.ServiceOpenStack.create_or_update', return_value=r)
    assert cmd.main(['service', 'create', '--format=json', 'name']) == 0
    out, err = capsys.readouterr()
    j = json.loads(out)
    assert {'fqdn': j['name']} == r



testinfra_hosts = ['ansible://icinga-host']


def test_bind(host):
    domain = host.run("hostname -d").stdout.strip()
    bind_host = host.get_host('ansible://bind-host',
                              ansible_inventory=host.backend.ansible_inventory)
    address = bind_host.ansible.get_variables()['ansible_host']
    for h in ('ns1', 'bind', 'bind-host'):
        cmd = host.run("getent hosts {}.{}".format(h, domain))
        print(cmd.stdout)
        print(cmd.stderr)
        assert 0 == cmd.rc
        assert address in cmd.stdout.strip()
        assert h + "." + domain in cmd.stdout.strip()
        # try also with shortnames
        cmd = host.run("getent hosts {}".format(h))
        assert 0 == cmd.rc
        assert address in cmd.stdout.strip()
        assert h + "." + domain in cmd.stdout.strip()


def test_recursion(host):
    cmd = host.run("getent hosts fsf.org")
    assert 0 == cmd.rc
    assert 'fsf.org' in cmd.stdout.strip()



import urllib3
import os
import testinfra
import time

import gitlab_utils
from enough.common.gitlab import GitLab

testinfra_hosts = ['ansible://gitlab-host']


def test_ci_runner(host, tmpdir):
    domain = host.run("hostname -d").stdout.strip()
    runner_host = testinfra.host.Host.get_host(
        'ansible://runner-host',
        ansible_inventory=host.backend.ansible_inventory)

    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
    gitlab = GitLab(f'https://lab.{domain}')
    gitlab.login('root', gitlab_utils.get_password())
    gitlab.recreate_project('root', 'testproject')
    runner_host.run("rm -f /tmp/*.out")
    os.system("""
    set -ex
    cd {directory}
    test -d testproject && exit 0
    mkdir testproject
    cd testproject
    git init
    git config user.email "you@example.com"
    git config user.name "Your Name"
    (
     echo 'jobs:'
     echo '  script: docker ps > /srv/DOCKER.out 2>&1'
    ) > .gitlab-ci.yml
    git add .gitlab-ci.yml
    git commit -m 'test'
    git config http.sslVerify false
    git remote add origin \
         https://root:{password}@{address}/root/testproject.git
    git push -u origin master
    """.format(password=gitlab_utils.get_password(),
               address=f'lab.{domain}',
               directory=str(tmpdir)))

    for (what, expected) in (('DOCKER', 'CONTAINER'),):
        success = False
        for _ in range(40):
            if (runner_host.file('/srv/' + what + '.out').exists and
                    runner_host.file('/srv/' + what + '.out').contains(
                        expected)):
                success = True
                break
            time.sleep(5)
        assert success


def test_docker_cleanup(host):
    runner_host = testinfra.host.Host.get_host(
        'ansible://runner-host',
        ansible_inventory=host.backend.ansible_inventory)

    with runner_host.sudo():
        cmd = runner_host.run("/etc/cron.daily/docker-cleanup")
        print(cmd.stdout)
        print(cmd.stderr)
        assert 0 == cmd.rc
        assert 'Total reclaimed space' in cmd.stdout


def test_notify_email(host):
    domain = host.run("hostname -d").stdout.strip()
    with host.sudo():
        cmd = host.run("systemctl stop postfix")
        command = """\
        docker exec --user git gitlab bin/rails runner -e production \
             "Notify.test_email('root', 'Message Subject', 'Message Body').deliver_now"
        """
        cmd = host.run(command)
        print(cmd.stdout)
        print('--------------------------')
        print(cmd.stderr)
        assert 1 == cmd.rc
        #
        # we're not testing if mail can actually be delivered, only if the
        # SMTP server configured in GitLab is the one we expected
        #
        assert (f'Connection refused - connect(2) for "gitlab-host.{domain}" '
                'port 25 (Errno::ECONNREFUSED)')



import logging
import os
import re
import sh
import shutil
import textwrap
import yaml

from enough.common.dotenough import DotEnoughOpenStack, DotEnough, Hosts
from enough.common.host import host_factory
from enough.common.service import service_factory
from enough.common.openstack import OpenStack, Heat
from enough.common import retry
from enough.common import ansible_utils

log = logging.getLogger(__name__)


class Enough(object):

    def __init__(self, config_dir, share_dir, **kwargs):
        self.config_dir = config_dir
        self.share_dir = share_dir
        self.init_args = kwargs.copy()
        self.set_args(**kwargs)

    def set_args(self, **kwargs):
        self.args = self.init_args.copy()
        self.args.update(kwargs)

        self.ansible = ansible_utils.Ansible(self.config_dir, self.share_dir,
                                             self.args.get('inventory'))
        if self.args.get('driver') == 'openstack':
            self.dotenough = DotEnoughOpenStack(self.config_dir, self.args['domain'])
            cloud = self.ansible.get_global_variable('openstack_cloud')
            if 'cloud' not in self.args:
                self.args['cloud'] = cloud
            self.dotenough.populate_provider(self.ansible, cloud)
        else:
            self.dotenough = DotEnough(self.config_dir, self.args['domain'])
        self.hosts = Hosts(self.config_dir)
        self.host = host_factory(self.config_dir, self.share_dir, **self.args)
        self.service = service_factory(self.config_dir, self.share_dir, **self.args)
        if self.args.get('driver') == 'openstack':
            self.openstack = OpenStack(self.config_dir, **self.args)
            self.heat = Heat(self.config_dir, **self.args)
        self.playbook = ansible_utils.Playbook(self.config_dir, self.share_dir,
                                               self.args.get('inventory'))

    def create_missings(self, hosts):
        public_key = f'{self.config_dir}/infrastructure_key.pub'
        r = self.heat.create_missings(hosts, public_key)
        self.update_internal_subnet_dns()
        return r

    def update_internal_subnet_dns(self):
        self.hosts.load()
        if not self.hosts.get_ip('bind-host'):
            log.info('do not update internal dns because bind-host does not exist')
            return
        log.info('update internal network dns server with bind-host IP address')
        dns = []
        for host in self.service.service2group['bind']:
            dns.append(self.openstack.server_ip_in_network(host, self.openstack.INTERNAL_NETWORK))
        self.openstack.subnet_update_dns(self.openstack.INTERNAL_NETWORK, *dns)

    def _vpn_credentials_path(self):
        return f'{self.config_dir}/openvpn/localhost.tar.gz'

    def vpn_has_credentials(self):
        return os.path.exists(self._vpn_credentials_path())

    def vpn_connect(self):
        sh.sudo.tar('-xf', self._vpn_credentials_path(),
                    _cwd='/etc/openvpn/client')

        network_prefix = self.ansible.get_variable('openstack_internal_network_prefix',
                                                   'bind-host')
        sh.sudo.openvpn(
            '--config', 'localhost.conf',
            '--writepid', '/var/run/openvpn-localhost.pid',
            _cwd='/etc/openvpn/client', _bg=True, _truncate_exc=False)

        @retry.retry(AssertionError, tries=5)
        def check_route():
            output = sh.ip.route()
            assert network_prefix in output, f'{network_prefix} in {output}'

        check_route()

    def vpn_disconnect(self):
        if os.path.exists('/var/run/openvpn-localhost.pid'):
            pid = open('/var/run/openvpn-localhost.pid').read().strip()
            sh.sudo.kill(pid)
            sh.sudo.rm('/var/run/openvpn-localhost.pid')

    @staticmethod
    def replace_production(clouds, target_cloud):
        content = yaml.safe_load(open(clouds))
        content['clouds']['production'] = content['clouds'][target_cloud]
        del content['clouds'][target_cloud]
        open(clouds, 'w').write(yaml.dump(content))

    def clone(self, target_domain, target_cloud, clobber_cloud):
        config_base = os.path.dirname(self.config_dir)
        config_dir = f'{config_base}/{target_domain}'
        if clobber_cloud and os.path.exists(config_dir):
            shutil.rmtree(config_dir)
        if not os.path.exists(config_dir):
            sh.rsync('-a', '--delete',
                     '--exclude=private-key.yml',
                     '--exclude=hosts.yml',
                     '--exclude=group_vars/all/domain.yml',
                     '--exclude=group_vars/all/production_domain.yml',
                     '--exclude=group_vars/all/internal_network.yml',
                     f'{self.config_dir}/', f'{config_dir}/')
            all_dir = f'{config_dir}/inventory/group_vars/all'
            open(f'{all_dir}/internal_network.yml', 'w').write(textwrap.dedent(f"""\
            ---
            openstack_internal_network_prefix: "10.11.10"
            openstack_internal_network_cidr: "10.11.10.0/24"
            """))
            clone_override_dir = f'{config_base}/clone-override'
            if os.path.exists(clone_override_dir):
                log.info(f'overriding with {clone_override_dir}')
                sh.rsync('-av', '--checksum', f'{clone_override_dir}/', f'{config_dir}/')
            if os.path.exists(f'{self.config_dir}.pass'):
                shutil.copyfile(f'{self.config_dir}.pass', f'{config_dir}.pass')
            self.replace_production(f'{all_dir}/clouds.yml', target_cloud)
        kwargs = {
            'driver': self.args['driver'],
            'domain': target_domain,
        }
        clone = Enough(config_dir, self.share_dir, **kwargs)
        return clone

    def create_copy_host(self, name, original_volume, copy_volume):
        self.set_args(name=name)
        h = self.host.create_or_update()
        if self.openstack.o.volume.list(
                '-c', 'Status', '--format=value', '--name', copy_volume).strip() == 'available':
            self.openstack.o.server.add.volume(name, copy_volume)
        self.playbook.run([
            f'--private-key={self.dotenough.private_key()}',
            '--extra-vars', f'encrypted_volume_name={original_volume}',
            '--limit', f'localhost,{name}',
            f'{self.share_dir}/copy-playbook.yml',
        ])
        return h['ipv4']

    def delete_copy_host(self, name):
        self.set_args(name=[name])
        self.host.delete()

    def _rsync_copy_host(self, from_ip, to_ip):
        sh.ssh('-i', self.dotenough.private_key(),
               '-o', 'StrictHostKeyChecking=no',
               '-o', 'UserKnownHostsFile=/dev/null',
               f'root@{from_ip}',
               'rsync', '-avHS', '--numeric-ids', '--delete',
               '/srv/', f'root@{to_ip}:/srv/',
               _tee=True,
               _out=lambda x: log.info(x.strip()),
               _err=lambda x: log.info(x.strip()),
               _truncate_exc=False)

    def clone_volume_from_snapshot(self, clone, snapshot):
        (from_ip, to_ip, from_volume, to_volume) = self._clone_volume_from_snapshot_body(
            clone, snapshot)
        self._rsync_copy_host(from_ip, to_ip)
        self._clone_volume_from_snapshot_cleanup(clone, from_volume)
        clone.openstack.o.volume.set('--name', snapshot, to_volume)

    def create_volume_from_snapshot(self, snapshot, volume):
        if not self.openstack.o.volume.list(
                '-c', 'Name', '--format=value',
                '--name', volume).strip() == volume:
            self.openstack.o.volume.create('--snapshot', snapshot, volume)

    def _clone_volume_from_snapshot_body(self, clone, snapshot):
        #
        #         in the current region
        # Step 1: create volume from snapshot
        #
        volume_name = self.volume_from_snapshot(snapshot)
        from_volume = f'copy-from-{volume_name}'
        self.create_volume_from_snapshot(snapshot, from_volume)

        size = self.openstack.o.volume.show(
            '-c', 'size', '--format=value', from_volume).strip()
        #
        #         in the current region
        # Step 2: attach the volume to a host dedicated to copying the content of the volume
        #
        from_ip = self.create_copy_host('copy-from-host', volume_name, from_volume)

        #
        #         in the clone region
        # Step 3: create an empty volume of the same size and attach it to a host dedicated
        #         to receive the copy
        #
        to_volume = f'copy-to-{volume_name}'
        if not clone.openstack.o.volume.list(
                '-c', 'Name', '--format=value',
                '--name', to_volume).strip() == to_volume:
            clone.openstack.o.volume.create('--size', size, to_volume)
        to_ip = clone.create_copy_host('copy-to-host', volume_name, to_volume)

        return (from_ip, to_ip, from_volume, to_volume)

    def _clone_volume_from_snapshot_cleanup(self, clone, copy_from_volume):
        self.delete_copy_host('copy-from-host')
        self.openstack.o.volume.delete(copy_from_volume)
        clone.delete_copy_host('copy-to-host')

    def destroy(self):
        self.openstack.destroy_everything(None)
        sh.rm('-r', self.config_dir)

    def cli_clone_volume(self):
        clone = self.clone(self.args['target_domain'],
                           self.args['target_cloud'],
                           self.args['clobber_cloud'])
        self.clone_volume_from_snapshot(clone, self.args['name'])

    @staticmethod
    def volume_from_snapshot(snapshot):
        return re.sub('^\d\d\d\d-\d\d-\d\d-', '', snapshot)

    def host_from_snapshot(self, snapshot):
        volume_name = self.volume_from_snapshot(snapshot)
        return self.heat.host_from_volume(volume_name)

    def create_service_matching_snapshot(self, snapshot):
        host = self.host_from_snapshot(snapshot)
        service = self.service.service_from_host(host)
        self.set_args(name=service, playbook='enough-playbook.yml')
        self.service.create_or_update()
        return host

    def restore(self):
        if self.args['target_domain']:
            return self.restore_remote(self.args['target_domain'],
                                       self.args['target_cloud'],
                                       self.args['clobber_cloud'],
                                       self.args['name'])
        else:
            return self.restore_local(self.args['name'])

    def restore_local(self, snapshot):
        host = self.host_from_snapshot(snapshot)
        volume = snapshot
        self.create_volume_from_snapshot(snapshot, volume)
        self.openstack.replace_volume(host, volume, delete_volume=False)
        return self

    def restore_remote(self, domain, target_cloud, clobber_cloud, snapshot):
        clone = self.clone(domain, target_cloud, clobber_cloud)
        self.clone_volume_from_snapshot(clone, snapshot)
        host = clone.create_service_matching_snapshot(snapshot)
        clone.openstack.replace_volume(host, snapshot, delete_volume=True)
        return clone

    class VolumeResizeUndefined(Exception):
        pass

    class VolumeResizeNoSize(Exception):
        pass

    def volume_resize(self, host, volume):
        d = self.heat.get_stack_definitions()[host]
        volumes = d.get('openstack_volumes')
        if not volumes:
            raise Enough.VolumeResizeUndefined(f'no openstack_volumes variable is set for {host}')
        volume_size = None
        for info in volumes:
            if info['name'] == volume:
                volume_size = info.get('size')
                break
        if volume_size is None:
            raise Enough.VolumeResizeNoSize(f'no size found for {volume} in the openstack_volumes '
                                            f'variable ({volumes}) for {host}')
        r = self.openstack.volume_resize(host, volume, int(volume_size))
        self.openstack.wait_for_ssh(d['ansible_host'], d['ansible_port'])
        return r



testinfra_hosts = ['ansible://website-host']


def test_retirement(host):
    with host.sudo():
        assert host.file("/srv/openvpn/easy-rsa/pki/issued/localhost.crt").exists
        assert not host.file("/srv/openvpn/easy-rsa/pki/issued/retired.crt").exists



import logging

testinfra_hosts = ['ansible://infrastructure1-host']
logger = logging.getLogger(__name__)


def test_internal_network(host):
    i2_host = host.get_host('ansible://infrastructure2-host',
                            ansible_inventory=host.backend.ansible_inventory)
    cmd = i2_host.run(r"hostname -I | sed -e 's/.*\(10.30.20.[0-9]*\).*/\1/'")
    logger.debug('stdout %s', cmd.stdout)
    logger.debug('stderr %s', cmd.stderr)
    assert 0 == cmd.rc
    ip = cmd.stdout.strip()
    assert '10.30.20' in ip

    cmd = host.run(f"""
    set -xe
    ping -c 1 {ip}
    """)
    logger.debug('stdout %s', cmd.stdout)
    logger.debug('stderr %s', cmd.stderr)
    assert 0 == cmd.rc
    assert '1 packets transmitted' in cmd.stdout



testinfra_hosts = ['ansible://bind-host']


def test_can_curl_locally(host):
    with host.sudo():
        host.run("apt-get install -y curl")
    assert host.run("curl -m 5 -I https://bind-host.$(hostname -d)").rc == 0


def test_ownca_can_curl(host):
    ownca_host = host.get_host('ansible://ownca-host',
                               ansible_inventory=host.backend.ansible_inventory)
    with ownca_host.sudo():
        ownca_host.run("apt-get install -y curl")
    assert ownca_host.run("curl -m 5 -I https://bind-host.$(hostname -d)").rc == 0
